from typing import Union

from ..asset import Asset
from ..call_builder.base_call_builder import BaseCallBuilder
from ..client.base_async_client import BaseAsyncClient
from ..client.base_sync_client import BaseSyncClient


class TradesCallBuilder(BaseCallBuilder):
    """ Creates a new :class:`TradesCallBuilder` pointed to server defined by horizon_url.
    Do not create this object directly, use :func:`stellar_sdk.server.Server.trades`.

    See `Trades <https://www.stellar.org/developers/horizon/reference/endpoints/trades.html)>`_

    :param horizon_url: Horizon server URL.
    :param client: The client instance used to send request.
    """

    def __init__(
        self, horizon_url: str, client: Union[BaseAsyncClient, BaseSyncClient]
    ) -> None:
        super().__init__(horizon_url, client)
        self.endpoint: str = "trades"

    def for_asset_pair(self, base: Asset, counter: Asset) -> "TradesCallBuilder":
        """Filter trades for a specific asset pair (orderbook)

        :param base: base asset
        :param counter: counter asset
        :return: current TradesCallBuilder instance
        """
        params = {
            "base_asset_type": base.type,
            "base_asset_code": None if base.is_native() else base.code,
            "base_asset_issuer": base.issuer,
            "counter_asset_type": counter.type,
            "counter_asset_code": None if counter.is_native() else counter.code,
            "counter_asset_issuer": counter.issuer,
        }
        self._add_query_params(params)
        return self

    def for_offer(self, offer_id: Union[int, str]) -> "TradesCallBuilder":
        """Filter trades for a specific offer

        See `Trades for Offer <https://www.stellar.org/developers/horizon/reference/endpoints/trades-for-offer.html>`_

        :param offer_id: offer id
        :return: current TradesCallBuilder instance
        """
        self.endpoint = f"offers/{offer_id}/trades"
        return self

    def for_account(self, account_id: str) -> "TradesCallBuilder":
        """Filter trades for a specific account

        See `Trades for Account <https://www.stellar.org/developers/horizon/reference/endpoints/trades-for-account.html>`_

        :param account_id: account id
        :return: current TradesCallBuilder instance
        """
        self.endpoint = f"accounts/{account_id}/trades"
        return self



from decimal import Decimal
from typing import Optional, Union

from ..asset import Asset
from ..exceptions import ValueError, TypeError
from ..keypair import Keypair
from ..price import Price

_LOWER_LIMIT = "0"
_UPPER_LIMIT = "922337203685.4775807"
_EXPONENT = 7


def check_source(source: Optional[str]) -> None:
    if source is not None:
        check_ed25519_public_key(source)


def check_ed25519_public_key(public_key: str) -> None:
    Keypair.from_public_key(public_key)


def check_asset_code(asset_code: str) -> None:
    Asset.check_if_asset_code_is_valid(asset_code)


def check_price(price: Union[str, Decimal, Price]) -> None:
    if not (
        isinstance(price, str) or isinstance(price, Decimal) or isinstance(price, Price)
    ):
        raise TypeError(f"price should be type of {str}, {Decimal} or {Price}.")

    if isinstance(price, str) or isinstance(price, Decimal):
        check_amount(price)


def check_amount(amount: Union[str, Decimal]) -> None:
    if not (isinstance(amount, str) or isinstance(amount, Decimal)):
        raise TypeError(f"amount should be type of {str} or {Decimal}.")
    amount = Decimal(amount)
    if abs(amount.as_tuple().exponent) > _EXPONENT:
        raise ValueError(
            f"Value of '{amount}' must have at most 7 digits after the decimal."
        )
    if amount < Decimal(_LOWER_LIMIT) or amount > Decimal(_UPPER_LIMIT):
        raise ValueError(
            f"Value of '{amount}' must represent a positive number "
            f"and the max valid value is {_UPPER_LIMIT}."
        )



from .account_merge import AccountMerge
from .allow_trust import AllowTrust, TrustLineEntryFlag
from .begin_sponsoring_future_reserves import BeginSponsoringFutureReserves
from .bump_sequence import BumpSequence
from .change_trust import ChangeTrust
from .claim_claimable_balance import ClaimClaimableBalance
from .create_account import CreateAccount
from .create_claimable_balance import CreateClaimableBalance, Claimant, ClaimPredicate
from .create_passive_sell_offer import CreatePassiveSellOffer
from .end_sponsoring_future_reserves import EndSponsoringFutureReserves
from .inflation import Inflation
from .manage_buy_offer import ManageBuyOffer
from .manage_data import ManageData
from .manage_sell_offer import ManageSellOffer
from .operation import Operation
from .path_payment import PathPayment
from .path_payment_strict_receive import PathPaymentStrictReceive
from .path_payment_strict_send import PathPaymentStrictSend
from .payment import Payment
from .revoke_sponsorship import RevokeSponsorship
from .set_options import SetOptions, Flag

__all__ = [
    "Operation",
    "AccountMerge",
    "AllowTrust",
    "BumpSequence",
    "ChangeTrust",
    "ClaimClaimableBalance",
    "CreateAccount",
    "CreateClaimableBalance",
    "Claimant",
    "ClaimPredicate",
    "CreatePassiveSellOffer",
    "Inflation",
    "ManageBuyOffer",
    "ManageData",
    "ManageSellOffer",
    "PathPayment",
    "PathPaymentStrictReceive",
    "PathPaymentStrictSend",
    "Payment",
    "SetOptions",
    "BeginSponsoringFutureReserves",
    "EndSponsoringFutureReserves",
    "RevokeSponsorship",
    "TrustLineEntryFlag",
    "Flag",  # TODO: act like TrustLineEntryFlag
]



import os
from typing import Any, Union

import nacl.signing as ed25519
from nacl.exceptions import BadSignatureError as NaclBadSignatureError

from .sep.mnemonic import Language, StellarMnemonic
from .exceptions import (
    BadSignatureError,
    MissingEd25519SecretSeedError,
    TypeError,
    AttributeError,
)
from .strkey import StrKey
from .xdr import Xdr

__all__ = ["Keypair"]


class Keypair:
    """The :class:`Keypair` object, which represents a signing and
    verifying key for use with the Stellar network.

    Instead of instantiating the class directly, we recommend using one of
    several class methods:

    * :meth:`Keypair.random`
    * :meth:`Keypair.from_secret`
    * :meth:`Keypair.from_public_key`

    :param verify_key: The verifying (public) Ed25519 key in the keypair.
    :param signing_key: The signing (private) Ed25519 key in the keypair.
    """

    def __init__(
        self, verify_key: ed25519.VerifyKey, signing_key: ed25519.SigningKey = None
    ) -> None:
        self.verify_key: ed25519.VerifyKey = _get_key_of_expected_type(
            verify_key, ed25519.VerifyKey
        )
        self.signing_key: ed25519.SigningKey = _get_key_of_expected_type(
            signing_key, ed25519.SigningKey
        )

    @classmethod
    def random(cls) -> "Keypair":
        """Generate a :class:`Keypair` object from a randomly generated seed.

        :return: A new :class:`Keypair` instance derived by the randomly seed.
        """
        seed = os.urandom(32)
        return cls.from_raw_ed25519_seed(seed)

    @classmethod
    def from_secret(cls, secret: str) -> "Keypair":
        """Generate a :class:`Keypair` object from a secret seed.

        :param secret: strkey ed25519 seed, for example: `SB2LHKBL24ITV2Y346BU46XPEL45BDAFOOJLZ6SESCJZ6V5JMP7D6G5X`
        :return: A new :class:`Keypair` instance derived by the secret.
        :raise: :exc:`Ed25519SecretSeedInvalidError <stellar_sdk.exceptions.Ed25519SecretSeedInvalidError>`:
            if ``secret`` is not a valid ed25519 secret seed.
        """
        raw_secret = StrKey.decode_ed25519_secret_seed(secret)
        return cls.from_raw_ed25519_seed(raw_secret)

    @classmethod
    def from_public_key(cls, public_key: str) -> "Keypair":
        """Generate a :class:`Keypair` object from a public key.

        :param public_key: strkey ed25519 public key,
            for example: `GATPGGOIE6VWADVKD3ER3IFO2IH6DTOA5G535ITB3TT66FZFSIZEAU2B`
        :return: A new :class:`Keypair` instance derived by the public key.
        :raise: :exc:`Ed25519PublicKeyInvalidError <stellar_sdk.exceptions.Ed25519PublicKeyInvalidError>`:
            if ``public_key`` is not a valid ed25519 public key.
        """
        key = StrKey.decode_ed25519_public_key(public_key)
        return cls.from_raw_ed25519_public_key(key)

    @classmethod
    def from_raw_ed25519_seed(cls, raw_seed: bytes) -> "Keypair":
        """Generate a :class:`Keypair` object from ed25519 secret key seed raw bytes.

        :param raw_seed: ed25519 secret key seed raw bytes
        :return: A new :class:`Keypair` instance derived by the ed25519 secret key seed raw bytes
        """
        signing_key = ed25519.SigningKey(raw_seed)
        verify_key = signing_key.verify_key
        return cls(verify_key, signing_key)

    @classmethod
    def from_raw_ed25519_public_key(cls, raw_public_key: bytes) -> "Keypair":
        """Generate a :class:`Keypair` object from ed25519 public key raw bytes.

        :param raw_public_key: ed25519 public key raw bytes
        :return: A new :class:`Keypair` instance derived by the ed25519 public key raw bytes
        """
        verify_key = ed25519.VerifyKey(raw_public_key)
        return cls(verify_key)

    @property
    def secret(self) -> str:
        """Returns secret key associated with this :class:`Keypair` instance

        :return: secret key
        :raise: :exc:`MissingEd25519SecretSeedError <stellar_sdk.exceptions.MissingEd25519SecretSeedError>`
            The :class:`Keypair` does not contain secret seed
        """
        if not self.signing_key:
            raise MissingEd25519SecretSeedError(
                "The keypair does not contain secret seed. Use Keypair.from_secret or "
                "Keypair.random to create a new keypair with a secret seed."
            )

        return StrKey.encode_ed25519_secret_seed(self.raw_secret_key())

    @secret.setter
    def secret(self, secret: str) -> None:
        raise AttributeError(
            "Please use `Keypair.from_secret` to generate a new Keypair object."
        )

    @property
    def public_key(self) -> str:
        """Returns public key associated with this :class:`Keypair` instance

        :return: public key
        """
        return StrKey.encode_ed25519_public_key(self.raw_public_key())

    @public_key.setter
    def public_key(self, public_key: str) -> None:
        raise AttributeError(
            "Please use `Keypair.from_public_key` to generate a new Keypair object."
        )

    def xdr_public_key(self) -> Xdr.types.PublicKey:
        """
        :return: xdr public key
        """
        return Xdr.types.PublicKey(Xdr.const.KEY_TYPE_ED25519, bytes(self.verify_key))

    def xdr_account_id(self) -> Xdr.types.PublicKey:
        return self.xdr_public_key()

    def xdr_muxed_account(self) -> Xdr.types.MuxedAccount:
        return Xdr.types.MuxedAccount(
            Xdr.const.KEY_TYPE_ED25519, bytes(self.verify_key)
        )

    def raw_public_key(self) -> bytes:
        """Returns raw public key.

        :return: raw public key
        """
        return bytes(self.verify_key)

    def signature_hint(self) -> bytes:
        """Returns signature hint associated with this :class:`Keypair` instance

        :return: signature hint
        """
        return bytes(self.xdr_account_id().ed25519[-4:])

    def raw_secret_key(self) -> bytes:
        """Returns raw secret key.

        :return: raw secret key
        """
        return bytes(self.signing_key)

    def can_sign(self) -> bool:
        """Returns `True` if this :class:`Keypair` object contains secret key and can sign.

        :return: `True` if this :class:`Keypair` object contains secret key and can sign
        """
        return self.signing_key is not None

    def sign(self, data: bytes) -> bytes:
        """Sign the provided data with the keypair's private key.

        :param data: The data to sign.
        :return: signed bytes
        :raise: :exc:`MissingEd25519SecretSeedError <stellar_sdk.exceptions.MissingEd25519SecretSeedError>`:
            if :class:`Keypair` does not contain secret seed.
        """
        if not self.can_sign():
            raise MissingEd25519SecretSeedError(
                "The keypair does not contain secret seed. Use Keypair.from_secret or "
                "Keypair.random to create a new keypair with a secret seed."
            )
        return self.signing_key.sign(data).signature

    def verify(self, data: bytes, signature: bytes) -> None:
        """Verify the provided data and signature match this keypair's public key.

        :param data: The data that was signed.
        :param signature: The signature.
        :raise: :exc:`BadSignatureError <stellar_sdk.exceptions.BadSignatureError>`:
            if the verification failed and the signature was incorrect.
        """
        try:
            self.verify_key.verify(data, signature)
        except NaclBadSignatureError:
            raise BadSignatureError("Signature verification failed.")

    @staticmethod
    def generate_mnemonic_phrase(
        language: Union[Language, str] = Language.ENGLISH, strength: int = 128
    ):
        """Generate a mnemonic phrase.

        :param language: The language of the mnemonic phrase, defaults to english.
        :param strength: The complexity of the mnemonic.
        :return: A mnemonic phrase.
        """
        mnemonic_phrase = StellarMnemonic(language).generate(strength)
        return mnemonic_phrase

    @classmethod
    def from_mnemonic_phrase(
        cls,
        mnemonic_phrase: str,
        language: Union[Language, str] = Language.ENGLISH,
        passphrase: str = "",
        index: int = 0,
    ):
        """Generate a :class:`Keypair` object via a mnemonic
        phrase.

        :param mnemonic_phrase: A unique string used to deterministically generate keypairs.
        :param language: The language of the mnemonic phrase, defaults to english.
        :param passphrase: An optional passphrase used as part of the salt
            during PBKDF2 rounds when generating the seed from the mnemonic.
        :param index: The index of the keypair generated by the mnemonic.
            This allows for multiple Keypairs to be derived from the same
            mnemonic, such as::

            >>> from stellar_sdk.keypair import Keypair
            >>> mnemonic = 'update hello cry airport drive chunk elite boat shaft sea describe number'  # Don't use this mnemonic in practice.
            >>> kp1 = Keypair.from_mnemonic_phrase(mnemonic, index=0)
            >>> kp2 = Keypair.from_mnemonic_phrase(mnemonic, index=1)
            >>> kp3 = Keypair.from_mnemonic_phrase(mnemonic, index=2)
        :return: A new :class:`Keypair` instance derived from the mnemonic.
        """
        raw_ed25519_seed = StellarMnemonic(language).to_seed(
            mnemonic_phrase, passphrase, index
        )
        return cls.from_raw_ed25519_seed(raw_ed25519_seed)

    def sign_decorated(self, data) -> Xdr.types.DecoratedSignature:
        """Sign the provided data with the keypair's private key and returns DecoratedSignature.

        :param data: signed bytes
        :return: sign decorated
        """
        signature = self.sign(data)
        hint = self.signature_hint()
        return Xdr.types.DecoratedSignature(hint, signature)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented  # pragma: no cover
        return (
            self.verify_key == other.verify_key
            and self.signing_key == other.signing_key
        )

    def __str__(self):
        return f"<Keypair [public_key={self.public_key}]>"


def _get_key_of_expected_type(key: Any, expected_type: Any) -> Any:
    if key is not None and not isinstance(key, expected_type):
        raise TypeError(
            f"The given key_type={type(key)} is not of type {expected_type}."
        )
    return key



from decimal import Decimal
from typing import Union

from .operation import Operation
from .utils import check_ed25519_public_key, check_amount
from ..keypair import Keypair
from ..strkey import StrKey
from ..xdr import Xdr


class CreateAccount(Operation):
    """The :class:`CreateAccount` object, which represents a Create Account
    operation on Stellar's network.

    This operation creates and funds a new account with the specified starting
    balance.

    Threshold: Medium

    :param destination: Destination account ID to create an account for.
    :param starting_balance: Amount in XLM the account should be
        funded for. Must be greater than the `reserve balance amount
        <https://www.stellar.org/developers/learn/concepts/fees.html>`_.
    :param source: The source account for the payment. Defaults to the
        transaction's source account.

    """

    def __init__(
        self,
        destination: str,
        starting_balance: Union[str, Decimal],
        source: str = None,
    ) -> None:
        super().__init__(source)
        check_ed25519_public_key(destination)
        check_amount(starting_balance)
        self.destination: str = destination
        self.starting_balance: Union[str, Decimal] = starting_balance

    @classmethod
    def type_code(cls) -> int:
        return Xdr.const.CREATE_ACCOUNT

    def _to_operation_body(self):
        destination = Keypair.from_public_key(self.destination).xdr_account_id()

        create_account_op = Xdr.types.CreateAccountOp(
            destination, Operation.to_xdr_amount(self.starting_balance)
        )

        body = Xdr.nullclass()
        body.type = Xdr.const.CREATE_ACCOUNT
        body.createAccountOp = create_account_op
        return body

    @classmethod
    def from_xdr_object(
        cls, operation_xdr_object: Xdr.types.Operation
    ) -> "CreateAccount":
        """Creates a :class:`CreateAccount` object from an XDR Operation object.

        """
        source = Operation.get_source_from_xdr_obj(operation_xdr_object)

        destination = StrKey.encode_ed25519_public_key(
            operation_xdr_object.body.createAccountOp.destination.ed25519
        )
        starting_balance = Operation.from_xdr_amount(
            operation_xdr_object.body.createAccountOp.startingBalance
        )

        op = cls(
            source=source, destination=destination, starting_balance=starting_balance
        )
        op._source_muxed = Operation.get_source_muxed_from_xdr_obj(operation_xdr_object)
        return op



from typing import Union

from ..call_builder.base_call_builder import BaseCallBuilder
from ..client.base_async_client import BaseAsyncClient
from ..client.base_sync_client import BaseSyncClient


class TransactionsCallBuilder(BaseCallBuilder):
    """ Creates a new :class:`TransactionsCallBuilder` pointed to server defined by horizon_url.
    Do not create this object directly, use :func:`stellar_sdk.server.Server.transactions`.

    See `All Transactions <https://www.stellar.org/developers/horizon/reference/endpoints/transactions-all.html>`_

    :param horizon_url: Horizon server URL.
    :param client: The client instance used to send request.
    """

    def __init__(
        self, horizon_url: str, client: Union[BaseAsyncClient, BaseSyncClient]
    ) -> None:
        super().__init__(horizon_url, client)
        self.endpoint: str = "transactions"

    def transaction(self, transaction_hash: str) -> "TransactionsCallBuilder":
        """The transaction details endpoint provides information on a single transaction.
        The transaction hash provided in the hash argument specifies which transaction to load.

        See `Transaction Details <https://www.stellar.org/developers/horizon/reference/endpoints/transactions-single.html>`_


        :param transaction_hash: transaction hash
        :return: current TransactionsCallBuilder instance
        """
        self.endpoint = f"transactions/{transaction_hash}"
        return self

    def for_account(self, account_id: str) -> "TransactionsCallBuilder":
        """This endpoint represents all transactions that affected a given account.

        See `Transactions for Account <https://www.stellar.org/developers/horizon/reference/endpoints/transactions-for-account.html>`_

        :param account_id: account id
        :return: current TransactionsCallBuilder instance
        """
        self.endpoint = f"accounts/{account_id}/transactions"
        return self

    def for_ledger(self, sequence: Union[str, int]) -> "TransactionsCallBuilder":
        """This endpoint represents all transactions in a given ledger.

        See `Transactions for Ledger <https://www.stellar.org/developers/horizon/reference/endpoints/transactions-for-ledger.html>`_

        :param sequence: ledger sequence
        :return: current TransactionsCallBuilder instance
        """
        self.endpoint = f"ledgers/{sequence}/transactions"
        return self

    def include_failed(self, include_failed: bool) -> "TransactionsCallBuilder":
        """Adds a parameter defining whether to include failed transactions. By default only
        transactions of successful transactions are returned.

        :param include_failed: Set to `True` to include failed transactions.
        :return: current TransactionsCallBuilder instance
        """
        self._add_query_param("include_failed", include_failed)
        return self



from decimal import Decimal
from typing import Union, Optional

from .operation import Operation
from .utils import check_amount, check_ed25519_public_key
from ..keypair import Keypair
from ..asset import Asset
from ..utils import parse_ed25519_account_id_from_muxed_account_xdr_object
from ..xdr import Xdr


class Payment(Operation):
    """The :class:`Payment` object, which represents a Payment operation on
    Stellar's network.

    Sends an amount in a specific asset to a destination account.

    Threshold: Medium

    :param destination: The destination account ID.
    :param asset: The asset to send.
    :param amount: The amount to send.
    :param source: The source account for the payment. Defaults to the
        transaction's source account.

    """

    def __init__(
        self,
        destination: str,
        asset: Asset,
        amount: Union[str, Decimal],
        source: str = None,
    ) -> None:
        super().__init__(source)
        check_amount(amount)
        check_ed25519_public_key(destination)
        self._destination: str = destination
        self._destination_muxed: Optional[Xdr.types.MuxedAccount] = None
        self.asset: Asset = asset
        self.amount: Union[str, Decimal] = amount

    @property
    def destination(self) -> str:
        return self._destination

    @destination.setter
    def destination(self, value: str):
        check_ed25519_public_key(value)
        self._destination_muxed = None
        self._destination = value

    @classmethod
    def type_code(cls) -> int:
        return Xdr.const.PAYMENT

    def _to_operation_body(self) -> Xdr.nullclass:
        asset = self.asset.to_xdr_object()
        if self._destination_muxed is not None:
            destination = self._destination_muxed
        else:
            destination = Keypair.from_public_key(self._destination).xdr_muxed_account()
        amount = Operation.to_xdr_amount(self.amount)
        payment_op = Xdr.types.PaymentOp(destination, asset, amount)
        body = Xdr.nullclass()
        body.type = Xdr.const.PAYMENT
        body.paymentOp = payment_op
        return body

    @classmethod
    def from_xdr_object(cls, operation_xdr_object: Xdr.types.Operation) -> "Payment":
        """Creates a :class:`Payment` object from an XDR Operation
        object.

        """
        source = Operation.get_source_from_xdr_obj(operation_xdr_object)

        destination = parse_ed25519_account_id_from_muxed_account_xdr_object(
            operation_xdr_object.body.paymentOp.destination
        )
        asset = Asset.from_xdr_object(operation_xdr_object.body.paymentOp.asset)
        amount = Operation.from_xdr_amount(operation_xdr_object.body.paymentOp.amount)

        op = cls(source=source, destination=destination, asset=asset, amount=amount)
        op._destination_muxed = operation_xdr_object.body.paymentOp.destination
        op._source_muxed = Operation.get_source_muxed_from_xdr_obj(operation_xdr_object)
        return op



import json
from typing import Generator, Union, Dict, Any, Tuple

import requests
from requests import Session, RequestException
from requests.adapters import HTTPAdapter, DEFAULT_POOLSIZE
from sseclient import SSEClient
from urllib3.exceptions import NewConnectionError
from urllib3.util import Retry

from . import defines
from ..__version__ import __version__
from ..client.base_sync_client import BaseSyncClient
from ..client.response import Response
from ..exceptions import ConnectionError

DEFAULT_NUM_RETRIES = 3
DEFAULT_BACKOFF_FACTOR = 0.5
USER_AGENT = f"py-stellar-sdk/{__version__}/RequestsClient"
IDENTIFICATION_HEADERS = {
    "X-Client-Name": "py-stellar-sdk",
    "X-Client-Version": __version__,
}

__all__ = ["RequestsClient"]


class RequestsClient(BaseSyncClient):
    """The :class:`RequestsClient` object is a synchronous http client,
    which represents the interface for making requests to a server instance.

    :param pool_size: persistent connection to Horizon and connection pool
    :param num_retries: configurable request retry functionality
    :param request_timeout: the timeout for all GET requests
    :param post_timeout: the timeout for all POST requests
    :param backoff_factor: a backoff factor to apply between attempts after the second try
    :param session: the request session
    :param stream_session: the stream request session
    """

    def __init__(
        self,
        pool_size: int = DEFAULT_POOLSIZE,
        num_retries: int = DEFAULT_NUM_RETRIES,
        request_timeout: int = defines.DEFAULT_GET_TIMEOUT_SECONDS,
        post_timeout: float = defines.DEFAULT_POST_TIMEOUT_SECONDS,
        backoff_factor: float = DEFAULT_BACKOFF_FACTOR,
        session: Session = None,
        stream_session: Session = None,
    ):
        self.pool_size: int = pool_size
        self.num_retries: int = num_retries
        self.request_timeout: int = request_timeout
        self.post_timeout: float = post_timeout
        self.backoff_factor: float = backoff_factor

        # adding 504 to the tuple of statuses to retry
        self.status_forcelist: Tuple[int] = tuple(Retry.RETRY_AFTER_STATUS_CODES) + (
            504,
        )

        # configure standard session

        # configure retry handler
        retry = Retry(
            total=self.num_retries,
            backoff_factor=self.backoff_factor,
            redirect=0,
            status_forcelist=self.status_forcelist,
            method_whitelist=frozenset(["GET", "POST"]),
            raise_on_status=False,
        )
        # init transport adapter
        adapter = HTTPAdapter(
            pool_connections=self.pool_size,
            pool_maxsize=self.pool_size,
            max_retries=retry,
        )

        headers = {**IDENTIFICATION_HEADERS, "User-Agent": USER_AGENT}

        # init session
        if session is None:
            session = requests.Session()

            # set default headers
            session.headers.update(headers)

            session.mount("http://", adapter)
            session.mount("https://", adapter)
        self._session: Session = session

        if stream_session is None:
            # configure SSE session (differs from our standard session)
            stream_session = requests.Session()

            sse_retry = Retry(
                total=1000000, redirect=0, status_forcelist=self.status_forcelist
            )
            sse_adapter = HTTPAdapter(
                pool_connections=self.pool_size,
                pool_maxsize=self.pool_size,
                max_retries=sse_retry,
            )

            stream_session.headers.update(headers)
            stream_session.mount("http://", sse_adapter)
            stream_session.mount("https://", sse_adapter)
        self._stream_session: Session = stream_session

    def get(self, url: str, params: Dict[str, str] = None) -> Response:
        """Perform HTTP GET request.

        :param url: the request url
        :param params: the request params
        :return: the response from server
        :raise: :exc:`ConnectionError <stellar_sdk.exceptions.ConnectionError>`
        """
        try:
            resp = self._session.get(url, params=params, timeout=self.request_timeout)
        except (RequestException, NewConnectionError) as err:
            raise ConnectionError(err)
        return Response(
            status_code=resp.status_code,
            text=resp.text,
            headers=dict(resp.headers),
            url=resp.url,
        )

    def post(self, url: str, data: Dict[str, str] = None) -> Response:
        """Perform HTTP POST request.

        :param url: the request url
        :param data: the data send to server
        :return: the response from server
        :raise: :exc:`ConnectionError <stellar_sdk.exceptions.ConnectionError>`
        """
        try:
            resp = self._session.post(url, data=data, timeout=self.post_timeout)
        except (RequestException, NewConnectionError) as err:
            raise ConnectionError(err)
        return Response(
            status_code=resp.status_code,
            text=resp.text,
            headers=dict(resp.headers),
            url=resp.url,
        )

    def stream(
        self, url: str, params: Dict[str, str] = None
    ) -> Generator[Dict[str, Any], None, None]:
        """Creates an EventSource that listens for incoming messages from the server.

        See `Horizon Response Format <https://www.stellar.org/developers/horizon/reference/responses.html>`_

        See `MDN EventSource <https://developer.mozilla.org/en-US/docs/Web/API/EventSource>`_

        :param url: the request url
        :param params: the request params
        :return: a Generator for server response
        :raise: :exc:`ConnectionError <stellar_sdk.exceptions.ConnectionError>`
        """
        query_params: Dict[str, Union[int, float, str]] = {**IDENTIFICATION_HEADERS}
        if params:
            query_params = {**params, **query_params}
        stream_client = _SSEClient(
            url,
            retry=0,
            session=self._stream_session,
            connect_retry=-1,
            params=query_params,
        )
        for message in stream_client:
            yield message

    def close(self) -> None:
        """Close underlying connector.

        Release all acquired resources.
        """
        self._session.close()
        self._stream_session.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class _SSEClient:
    def __init__(
        self,
        url: str,
        last_id: Union[str, int] = None,
        retry: int = 3000,
        session: Session = None,
        chunk_size: int = 1024,
        connect_retry: int = 0,
        **kwargs,
    ):
        if SSEClient is None:
            raise ImportError(
                "SSE not supported, missing `stellar-base-sseclient` module"
            )  # pragma: no cover

        self.client = SSEClient(
            url, last_id, retry, session, chunk_size, connect_retry, **kwargs
        )

    def __iter__(self):
        return self

    def __next__(self) -> Dict[str, Any]:
        while True:
            msg = next(self.client)
            data = msg.data
            if data != '"hello"' and data != '"byebye"':
                return json.loads(data)



import os
from urllib.parse import urljoin

import requests

BASE_XDR_GITHUB_URL = (
    "https://raw.githubusercontent.com/stellar/stellar-core/master/src/xdr/"
)
XDR_FILES = (
    "Stellar-SCP.x",
    "Stellar-ledger-entries.x",
    "Stellar-ledger.x",
    "Stellar-overlay.x",
    "Stellar-transaction.x",
    "Stellar-types.x",
)
BASE_DIR = os.path.abspath(os.path.join(__file__, "../../../.xdr"))
print(f"Downloading xdr files from {BASE_XDR_GITHUB_URL}")
for filename in XDR_FILES:
    print(f"Downloading {filename}")
    url = urljoin(BASE_XDR_GITHUB_URL, filename)
    file = os.path.join(BASE_DIR, filename)
    resp = requests.get(url, allow_redirects=True)
    open(file, "wb").write(resp.content)
print("Finished")



from typing import Union

from ..asset import Asset
from ..call_builder.base_call_builder import BaseCallBuilder
from ..client.base_async_client import BaseAsyncClient
from ..client.base_sync_client import BaseSyncClient
from ..exceptions import ValueError


class TradeAggregationsCallBuilder(BaseCallBuilder):
    """ Creates a new :class:`TradeAggregationsCallBuilder` pointed to server defined by horizon_url.
    Do not create this object directly, use :func:`stellar_sdk.server.Server.trade_aggregations`.

    Trade Aggregations facilitate efficient gathering of historical trade data.

    See `Trade Aggregations <https://www.stellar.org/developers/horizon/reference/endpoints/trade_aggregations.html>`_

    :param horizon_url: Horizon server URL.
    :param client: The client instance used to send request.
    :param base: base asset
    :param counter: counter asset
    :param resolution: segment duration as millis since epoch. *Supported values
        are 1 minute (60000), 5 minutes (300000), 15 minutes (900000),
        1 hour (3600000), 1 day (86400000) and 1 week (604800000).*
    :param start_time: lower time boundary represented as millis since epoch
    :param end_time: upper time boundary represented as millis since epoch
    :param offset: segments can be offset using this parameter.
        Expressed in milliseconds. *Can only be used if the resolution is greater than 1 hour.
        Value must be in whole hours, less than the provided resolution, and less than 24 hours.*
    """

    def __init__(
        self,
        horizon_url: str,
        client: Union[BaseAsyncClient, BaseSyncClient],
        base: Asset,
        counter: Asset,
        resolution: int,
        start_time: int = None,
        end_time: int = None,
        offset: int = None,
    ) -> None:
        super().__init__(horizon_url, client)
        self.endpoint: str = "trade_aggregations"

        if not self.__is_valid_resolution(resolution):
            raise ValueError(f"Invalid resolution: {resolution}")

        if offset and not self.__is_valid_offset(offset, resolution):
            raise ValueError(f"Invalid offset: {offset}")

        params = {
            "base_asset_type": base.type,
            "base_asset_code": None if base.is_native() else base.code,
            "base_asset_issuer": base.issuer,
            "counter_asset_type": counter.type,
            "counter_asset_code": None if counter.is_native() else counter.code,
            "counter_asset_issuer": counter.issuer,
            "start_time": start_time,
            "end_time": end_time,
            "resolution": resolution,
            "offset": offset,
        }
        self._add_query_params(params)

    def __is_valid_offset(self, offset: int, resolution: int) -> bool:
        """
        :param offset: Time offset in milliseconds
        :param resolution: Trade data resolution in milliseconds
        :return: `True` if the offset is allowed
        """
        hour = 3600000
        invalid = offset > resolution or offset >= 24 * hour or offset % hour != 0
        return not invalid

    def __is_valid_resolution(self, resolution: int) -> bool:
        """
        :param resolution: Trade data resolution in milliseconds
        :return: `True` if the resolution is allowed
        """
        allowed_resolutions = (60000, 300000, 900000, 3600000, 86400000, 604800000)
        if resolution in allowed_resolutions:
            return True
        return False



from typing import List

from .base_transaction_envelope import BaseTransactionEnvelope
from .fee_bump_transaction import FeeBumpTransaction
from .xdr import Xdr

__all__ = ["FeeBumpTransactionEnvelope"]


class FeeBumpTransactionEnvelope(BaseTransactionEnvelope["FeeBumpTransactionEnvelope"]):
    """The :class:`FeeBumpTransactionEnvelope` object, which represents a transaction
    envelope ready to sign and submit to send over the network.

    When a transaction is ready to be prepared for sending over the network, it
    must be put into a :class:`FeeBumpTransactionEnvelope`, which includes additional
    metadata such as the signers for a given transaction. Ultimately, this
    class handles signing and conversion to and from XDR for usage on Stellar's
    network.

    :param transaction: The fee bump transaction that is encapsulated in this envelope.
    :param list signatures: which contains a list of signatures that have
          already been created.
    :param network_passphrase: The network to connect to for verifying and retrieving additional attributes from.
    """

    def __init__(
        self,
        transaction: FeeBumpTransaction,
        network_passphrase: str,
        signatures: List[Xdr.types.DecoratedSignature] = None,
    ) -> None:
        super().__init__(network_passphrase, signatures)
        self.transaction = transaction

    def signature_base(self) -> bytes:
        """Get the signature base of this transaction envelope.

        Return the "signature base" of this transaction, which is the value
        that, when hashed, should be signed to create a signature that
        validators on the Stellar Network will accept.

        It is composed of a 4 prefix bytes followed by the xdr-encoded form of
        this transaction.

        :return: The signature base of this transaction envelope.

        """
        network_id = self.network_id
        tx_type = Xdr.StellarXDRPacker()
        tx_packer = Xdr.StellarXDRPacker()
        tx = self.transaction
        tx_type.pack_EnvelopeType(Xdr.const.ENVELOPE_TYPE_TX_FEE_BUMP)
        tx_packer.pack_FeeBumpTransaction(tx.to_xdr_object())
        tx_type_buffer = tx_type.get_buffer()
        tx_buffer = tx_packer.get_buffer()
        return network_id + tx_type_buffer + tx_buffer

    def to_xdr_object(self) -> Xdr.types.TransactionEnvelope:
        """Get an XDR object representation of this :class:`TransactionEnvelope`.

        :return: XDR TransactionEnvelope object
        """
        tx = self.transaction.to_xdr_object()
        te_type = Xdr.const.ENVELOPE_TYPE_TX_FEE_BUMP
        tx_envelope = Xdr.types.FeeBumpTransactionEnvelope(tx, self.signatures)
        return Xdr.types.TransactionEnvelope(type=te_type, feeBump=tx_envelope)

    @classmethod
    def from_xdr_object(
        cls, te_xdr_object: Xdr.types.TransactionEnvelope, network_passphrase: str
    ) -> "FeeBumpTransactionEnvelope":
        """Create a new :class:`FeeBumpTransactionEnvelope` from an XDR object.

        :param te_xdr_object: The XDR object that represents a fee bump transaction envelope.
        :param network_passphrase: The network to connect to for verifying and retrieving additional attributes from.
        :return: A new :class:`FeeBumpTransactionEnvelope` object from the given XDR TransactionEnvelope object.
        """
        te_type = te_xdr_object.type
        if te_type == Xdr.const.ENVELOPE_TYPE_TX_FEE_BUMP:
            tx = FeeBumpTransaction.from_xdr_object(
                te_xdr_object.feeBump, network_passphrase
            )
        else:
            raise ValueError(f"Invalid EnvelopeType: {te_xdr_object.type}.")
        signatures = te_xdr_object.signatures
        te = cls(tx, network_passphrase=network_passphrase, signatures=signatures)
        return te



from .operation import Operation
from ..xdr import Xdr


class BumpSequence(Operation):
    """The :class:`BumpSequence` object, which represents a
    BumpSequence operation on Stellar's network.

    Bump sequence allows to bump forward the sequence number of the source account of the
    operation, allowing to invalidate any transactions with a smaller sequence number.
    If the specified bumpTo sequence number is greater than the source account’s sequence number,
    the account’s sequence number is updated with that value, otherwise it’s not modified.

    Threshold: Low

    :param bump_to: Sequence number to bump to.
    :param source: The optional source account.

    """

    def __init__(self, bump_to: int, source: str = None) -> None:
        super().__init__(source)
        self.bump_to: int = bump_to

    @classmethod
    def type_code(cls) -> int:
        return Xdr.const.BUMP_SEQUENCE

    def _to_operation_body(self) -> Xdr.nullclass:
        bump_sequence_op = Xdr.types.BumpSequenceOp(self.bump_to)
        body = Xdr.nullclass()
        body.type = Xdr.const.BUMP_SEQUENCE
        body.bumpSequenceOp = bump_sequence_op
        return body

    @classmethod
    def from_xdr_object(
        cls, operation_xdr_object: Xdr.types.Operation
    ) -> "BumpSequence":
        """Creates a :class:`BumpSequence` object from an XDR Operation
        object.

        """
        source = Operation.get_source_from_xdr_obj(operation_xdr_object)

        bump_to = operation_xdr_object.body.bumpSequenceOp.bumpTo
        op = cls(source=source, bump_to=bump_to)
        op._source_muxed = Operation.get_source_muxed_from_xdr_obj(operation_xdr_object)
        return op



import warnings
from typing import Union

from ..asset import Asset
from ..call_builder.base_call_builder import BaseCallBuilder
from ..client.base_async_client import BaseAsyncClient
from ..client.base_sync_client import BaseSyncClient


class PathsCallBuilder(BaseCallBuilder):
    """ Creates a new :class:`PathsCallBuilder` pointed to server defined by horizon_url.
    Do not create this object directly, use :func:`stellar_sdk.server.Server.paths`.

    The Stellar Network allows payments to be made across assets through path payments. A path payment specifies a
    series of assets to route a payment through, from source asset (the asset debited from the payer) to destination
    asset (the asset credited to the payee).

    See `Find Payment Paths <https://www.stellar.org/developers/horizon/reference/endpoints/path-finding.html>`_
    
    :param horizon_url: Horizon server URL.
    :param client: The client instance used to send request.
    :param source_account: The sender's account ID. Any returned path must use a source that the sender can hold.
    :param destination_account: The destination account ID that any returned path should use.
    :param destination_asset: The destination asset.
    :param destination_amount: The amount, denominated in the destination asset, that any returned path should be able to satisfy.
    """

    def __init__(
        self,
        horizon_url: str,
        client: Union[BaseAsyncClient, BaseSyncClient],
        source_account: str,
        destination_account: str,
        destination_asset: Asset,
        destination_amount: str,
    ) -> None:

        warnings.warn(
            "Will be removed in version v3.0.0, "
            "use stellar_sdk.call_builder.StrictReceivePathsCallBuilder",
            DeprecationWarning,
        )

        super().__init__(horizon_url, client)
        self.endpoint: str = "paths"
        params = {
            "destination_account": destination_account,
            "source_account": source_account,
            "destination_amount": destination_amount,
            "destination_asset_type": destination_asset.type,
            "destination_asset_code": None
            if destination_asset.is_native()
            else destination_asset.code,
            "destination_asset_issuer": destination_asset.issuer,
        }
        self._add_query_params(params)



from typing import Union

from ..asset import Asset
from ..call_builder.base_call_builder import BaseCallBuilder
from ..client.base_async_client import BaseAsyncClient
from ..client.base_sync_client import BaseSyncClient
from ..utils import convert_assets_to_horizon_param


class ClaimableBalancesCallBuilder(BaseCallBuilder):
    """ Creates a new :class:`ClaimableBalancesCallBuilder` pointed to server defined by horizon_url.
    Do not create this object directly, use :func:`stellar_sdk.server.Server.claimable_balance`.

    :param horizon_url: Horizon server URL.
    :param client: The client instance used to send request.
    """

    def __init__(
        self, horizon_url, client: Union[BaseAsyncClient, BaseSyncClient]
    ) -> None:
        super().__init__(horizon_url, client)
        self.endpoint: str = "claimable_balances"

    def claimable_balance(
        self, claimable_balance_id: str
    ) -> "ClaimableBalancesCallBuilder":
        """Returns information and links relating to a single claimable balance.

        See `Claimable Balances <https://developers.stellar.org/api/resources/claimablebalances/list/>`_

        :param claimable_balance_id: claimable balance id
        :return: current AccountCallBuilder instance
        """
        self.endpoint = f"claimable_balances/{claimable_balance_id}"
        return self

    def for_sponsor(self, sponsor: str) -> "ClaimableBalancesCallBuilder":
        """Returns all claimable balances which are sponsored by the given account ID.

        See `Claimable Balances <https://developers.stellar.org/api/resources/claimablebalances/list/>`_

        :param sponsor: the sponsor id, for example: `GDGQVOKHW4VEJRU2TETD6DBRKEO5ERCNF353LW5WBFW3JJWQ2BRQ6KDD`
        :return: current ClaimableBalancesCallBuilder instance
        """
        self._add_query_param("sponsor", sponsor)
        return self

    def for_asset(self, asset: Asset) -> "ClaimableBalancesCallBuilder":
        """Returns all claimable balances which provide a balance for the given asset.

        See `Account Details <https://www.stellar.org/developers/horizon/reference/endpoints/accounts-single.html>`_

        :param asset: an asset
        :return: current ClaimableBalancesCallBuilder instance
        """
        assets_param = convert_assets_to_horizon_param([asset])
        self._add_query_param("asset", assets_param)
        return self

    def for_claimant(self, claimant: str) -> "ClaimableBalancesCallBuilder":
        """Returns all claimable balances which can be claimed by the given account ID.

        See `Account Details <https://www.stellar.org/developers/horizon/reference/endpoints/accounts-single.html>`_

        :param claimant: the account id, for example: `GDGQVOKHW4VEJRU2TETD6DBRKEO5ERCNF353LW5WBFW3JJWQ2BRQ6KDD`
        :return: current ClaimableBalancesCallBuilder instance
        """
        self._add_query_param("claimant", claimant)
        return self



from .__version__ import (
    __title__,
    __description__,
    __url__,
    __version__,
    __author__,
    __author_email__,
    __license__,
)

from .operation import *
from .operation import __all__ as operation_all

from .account import Account
from .asset import Asset
from .fee_bump_transaction import FeeBumpTransaction
from .fee_bump_transaction_envelope import FeeBumpTransactionEnvelope
from .keypair import Keypair
from .memo import Memo, NoneMemo, TextMemo, IdMemo, HashMemo, ReturnHashMemo
from .network import Network
from .price import Price
from .server import Server
from .signer import Signer
from .signer_key import SignerKey
from .time_bounds import TimeBounds
from .transaction import Transaction
from .transaction_builder import TransactionBuilder
from .transaction_envelope import TransactionEnvelope
from .client.requests_client import RequestsClient
from .client.aiohttp_client import AiohttpClient
from .helpers import *

__all__ = (
    [
        "__title__",
        "__description__",
        "__url__",
        "__version__",
        "__author__",
        "__author_email__",
        "__license__",
        "Account",
        "Asset",
        "FeeBumpTransaction",
        "FeeBumpTransactionEnvelope",
        "Keypair",
        "Memo",
        "NoneMemo",
        "TextMemo",
        "IdMemo",
        "HashMemo",
        "ReturnHashMemo",
        "Network",
        "Price",
        "Server",
        "Signer",
        "SignerKey",
        "TimeBounds",
        "Transaction",
        "TransactionBuilder",
        "TransactionEnvelope",
        "RequestsClient",
        "AiohttpClient",
    ]
    + operation_all
    + helpers.__all__
)



from .xdr import Xdr
from .exceptions import ValueError

__all__ = ["TimeBounds"]


class TimeBounds:
    """TimeBounds represents the time interval that a transaction is valid.

    The UNIX timestamp (in seconds), determined by ledger time,
    of a lower and upper bound of when this transaction will be valid.
    If a transaction is submitted too early or too late,
    it will fail to make it into the transaction set.
    **max_time** equal 0 means that it’s not set.

    See `Stellar's documentation on Transactions
    <https://www.stellar.org/developers/guides/concepts/transactions.html#time-bounds>`__
    for more information on how TimeBounds are used within transactions.


    :param min_time: the UNIX timestamp (in seconds)
    :param max_time: the UNIX timestamp (in seconds)
    :raises: :exc:`ValueError <stellar_sdk.exceptions.ValueError>`: if ``max_time`` less than ``min_time``.
    """

    def __init__(self, min_time: int, max_time: int) -> None:
        if min_time < 0:
            raise ValueError("min_time cannot be negative.")

        if max_time < 0:
            raise ValueError("max_time cannot be negative.")

        if 0 < max_time < min_time:
            raise ValueError("max_time must be >= min_time.")

        self.min_time: int = min_time
        self.max_time: int = max_time

    def to_xdr_object(self) -> Xdr.types.TimeBounds:
        """Returns the xdr object for this TimeBounds object.

        :return: XDR TimeBounds object
        """
        return Xdr.types.TimeBounds(minTime=self.min_time, maxTime=self.max_time)

    @classmethod
    def from_xdr_object(
        cls, time_bounds_xdr_object: Xdr.types.TimeBounds
    ) -> "TimeBounds":
        """Create a :class:`TimeBounds` from an XDR TimeBounds object.

        :param time_bounds_xdr_object: The XDR TimeBounds object.
        :return: A new :class:`TimeBounds` object from the given XDR TimeBounds object.
        """
        return cls(
            min_time=time_bounds_xdr_object.minTime,
            max_time=time_bounds_xdr_object.maxTime,
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented  # pragma: no cover
        return self.min_time == other.min_time and self.max_time == other.max_time

    def __str__(self):
        return f"<TimeBounds [min_time={self.min_time}, max_time={self.max_time}]>"



import warnings
from typing import Union

from ..asset import Asset
from ..call_builder.base_call_builder import BaseCallBuilder
from ..client.base_async_client import BaseAsyncClient
from ..client.base_sync_client import BaseSyncClient
from ..utils import convert_assets_to_horizon_param


class AccountsCallBuilder(BaseCallBuilder):
    """ Creates a new :class:`AccountsCallBuilder` pointed to server defined by horizon_url.
    Do not create this object directly, use :func:`stellar_sdk.server.Server.accounts`.

    :param horizon_url: Horizon server URL.
    :param client: The client instance used to send request.
    """

    def __init__(
        self, horizon_url, client: Union[BaseAsyncClient, BaseSyncClient]
    ) -> None:
        super().__init__(horizon_url, client)
        self.endpoint: str = "accounts"

    def account_id(self, account_id: str) -> "AccountsCallBuilder":
        """Returns information and links relating to a single account.
        The balances section in the returned JSON will also list all the trust lines this account has set up.

        See `Account Details <https://www.stellar.org/developers/horizon/reference/endpoints/accounts-single.html>`_

        :param account_id: account id, for example: `GDGQVOKHW4VEJRU2TETD6DBRKEO5ERCNF353LW5WBFW3JJWQ2BRQ6KDD`
        :return: current AccountCallBuilder instance
        """
        self.endpoint = f"accounts/{account_id}"
        return self

    def signer(self, signer: str) -> "AccountsCallBuilder":
        """Filtering accounts who have a given signer. The result is a list of accounts.

        See `Account Details <https://www.stellar.org/developers/horizon/reference/endpoints/accounts-single.html>`_

        :param signer: signer's account id, for example: `GDGQVOKHW4VEJRU2TETD6DBRKEO5ERCNF353LW5WBFW3JJWQ2BRQ6KDD`
        :return: current AccountCallBuilder instance
        """
        warnings.warn(
            "Will be removed in future, use AccountsCallBuilder.for_signer",
            DeprecationWarning,
        )
        return self.for_signer(signer)

    def for_signer(self, signer: str) -> "AccountsCallBuilder":
        """Filtering accounts who have a given signer. The result is a list of accounts.

        See `Account Details <https://www.stellar.org/developers/horizon/reference/endpoints/accounts-single.html>`_

        :param signer: signer's account id, for example: `GDGQVOKHW4VEJRU2TETD6DBRKEO5ERCNF353LW5WBFW3JJWQ2BRQ6KDD`
        :return: current AccountCallBuilder instance
        """
        self._add_query_param("signer", signer)
        return self

    def asset(self, asset: Asset) -> "AccountsCallBuilder":
        """Filtering accounts who have a trustline to an asset. The result is a list of accounts.

        See `Account Details <https://www.stellar.org/developers/horizon/reference/endpoints/accounts-single.html>`_

        :param asset: an issued asset
        :return: current AccountCallBuilder instance
        """
        warnings.warn(
            "Will be removed in future, use AccountsCallBuilder.for_asset",
            DeprecationWarning,
        )
        return self.for_asset(asset)

    def for_asset(self, asset: Asset) -> "AccountsCallBuilder":
        """Filtering accounts who have a trustline to an asset. The result is a list of accounts.

        See `Account Details <https://www.stellar.org/developers/horizon/reference/endpoints/accounts-single.html>`_

        :param asset: an issued asset
        :return: current AccountCallBuilder instance
        """
        assets_param = convert_assets_to_horizon_param([asset])
        self._add_query_param("asset", assets_param)
        return self

    def for_sponsor(self, sponsor: str) -> "AccountsCallBuilder":
        """Filtering accounts where the given account is sponsoring the account or any of its sub-entries.

        See `Account Details <https://developers.stellar.org/api/resources/accounts/single/>`_

        :param sponsor: the sponsor id, for example: `GDGQVOKHW4VEJRU2TETD6DBRKEO5ERCNF353LW5WBFW3JJWQ2BRQ6KDD`
        :return: current AccountCallBuilder instance
        """
        self._add_query_param("sponsor", sponsor)
        return self



from enum import IntFlag
from typing import List, Optional, Union

from .operation import Operation
from .utils import check_ed25519_public_key
from ..keypair import Keypair
from ..signer import Signer
from ..strkey import StrKey
from ..utils import pack_xdr_array, unpack_xdr_array
from ..xdr import Xdr

__all__ = ["Flag", "SetOptions"]


class Flag(IntFlag):
    """Indicates which flags to set. For details about the flags,
    please refer to the `accounts doc <https://www.stellar.org/developers/guides/concepts/accounts.html>`_.
    The bit mask integer adds onto the existing flags of the account.
    """

    AUTHORIZATION_REQUIRED = 1
    AUTHORIZATION_REVOCABLE = 2
    AUTHORIZATION_IMMUTABLE = 4


class SetOptions(Operation):
    """The :class:`SetOptions` object, which represents a SetOptions operation
    on Stellar's network.

    This operation sets the options for an account.

    For more information on the signing options, please refer to the `multi-sig
    doc <https://www.stellar.org/developers/guides/concepts/multi-sig.html>`_.

    When updating signers or other thresholds, the threshold of this operation
    is high.

    Threshold: Medium or High

    :param inflation_dest: Account of the inflation destination.
    :param clear_flags: Indicates which flags to clear. For details about the flags,
        please refer to the `accounts doc <https://www.stellar.org/developers/guides/concepts/accounts.html>`_.
        The `bit mask <https://en.wikipedia.org/wiki/Bit_field>`_ integer subtracts from the existing flags of the account.
        This allows for setting specific bits without knowledge of existing flags, you can also use
        :class:`stellar_sdk.operation.set_options.Flag`
        - AUTHORIZATION_REQUIRED = 1
        - AUTHORIZATION_REVOCABLE = 2
        - AUTHORIZATION_IMMUTABLE = 4
    :param set_flags: Indicates which flags to set. For details about the flags,
        please refer to the `accounts doc <https://www.stellar.org/developers/guides/concepts/accounts.html>`_.
        The bit mask integer adds onto the existing flags of the account.
        This allows for setting specific bits without knowledge of existing flags, you can also use
        :class:`stellar_sdk.operation.set_options.Flag`
        - AUTHORIZATION_REQUIRED = 1
        - AUTHORIZATION_REVOCABLE = 2
        - AUTHORIZATION_IMMUTABLE = 4
    :param master_weight: A number from 0-255 (inclusive) representing the weight of the master key.
        If the weight of the master key is updated to 0, it is effectively disabled.
    :param low_threshold: A number from 0-255 (inclusive) representing the threshold this account sets on all
        operations it performs that have `a low threshold <https://www.stellar.org/developers/guides/concepts/multi-sig.html>`_.
    :param med_threshold: A number from 0-255 (inclusive) representing the threshold this account sets on all
        operations it performs that have `a medium threshold <https://www.stellar.org/developers/guides/concepts/multi-sig.html>`_.
    :param high_threshold: A number from 0-255 (inclusive) representing the threshold this account sets on all
        operations it performs that have `a high threshold <https://www.stellar.org/developers/guides/concepts/multi-sig.html>`_.
    :param home_domain: sets the home domain used for
        reverse `federation <https://www.stellar.org/developers/guides/concepts/federation.html>`_ lookup.
    :param signer: Add, update, or remove a signer from the account.
    :param source: The source account (defaults to transaction source).

    """

    def __init__(
        self,
        inflation_dest: str = None,
        clear_flags: Union[int, Flag] = None,
        set_flags: Union[int, Flag] = None,
        master_weight: int = None,
        low_threshold: int = None,
        med_threshold: int = None,
        high_threshold: int = None,
        signer: Signer = None,
        home_domain: str = None,
        source: str = None,
    ) -> None:
        super().__init__(source)
        if inflation_dest is not None:
            check_ed25519_public_key(inflation_dest)

        if isinstance(set_flags, Flag):
            set_flags = set_flags.value

        if isinstance(clear_flags, Flag):
            clear_flags = clear_flags.value

        self.inflation_dest: str = inflation_dest
        self.clear_flags: int = clear_flags
        self.set_flags: int = set_flags
        self.master_weight: int = master_weight
        self.low_threshold: int = low_threshold
        self.med_threshold: int = med_threshold
        self.high_threshold: int = high_threshold
        self.home_domain: str = home_domain
        self.signer: Optional[Signer] = signer

    @classmethod
    def type_code(cls) -> int:
        return Xdr.const.SET_OPTIONS

    def _to_operation_body(self) -> Xdr.nullclass:
        if self.inflation_dest is not None:
            inflation_dest = [
                Keypair.from_public_key(self.inflation_dest).xdr_account_id()
            ]
        else:
            inflation_dest = []

        home_domain: List[bytes] = []
        if self.home_domain:
            home_domain = pack_xdr_array(bytes(self.home_domain, encoding="utf-8"))

        clear_flags = pack_xdr_array(self.clear_flags)
        set_flags = pack_xdr_array(self.set_flags)
        master_weight = pack_xdr_array(self.master_weight)
        low_threshold = pack_xdr_array(self.low_threshold)
        med_threshold = pack_xdr_array(self.med_threshold)
        high_threshold = pack_xdr_array(self.high_threshold)

        signer: List[Xdr.types.Signer] = []
        if self.signer:
            signer = [self.signer.to_xdr_object()]

        set_options_op = Xdr.types.SetOptionsOp(
            inflation_dest,
            clear_flags,
            set_flags,
            master_weight,
            low_threshold,
            med_threshold,
            high_threshold,
            home_domain,
            signer,
        )
        body = Xdr.nullclass()
        body.type = Xdr.const.SET_OPTIONS
        body.setOptionsOp = set_options_op
        return body

    @classmethod
    def from_xdr_object(cls, operation_xdr_object) -> "SetOptions":
        """Creates a :class:`SetOptions` object from an XDR Operation
        object.

        """
        source = Operation.get_source_from_xdr_obj(operation_xdr_object)

        inflation_dest = None
        if operation_xdr_object.body.setOptionsOp.inflationDest:
            inflation_dest = StrKey.encode_ed25519_public_key(
                operation_xdr_object.body.setOptionsOp.inflationDest[0].ed25519
            )

        clear_flags = unpack_xdr_array(
            operation_xdr_object.body.setOptionsOp.clearFlags
        )  # list
        set_flags = unpack_xdr_array(operation_xdr_object.body.setOptionsOp.setFlags)
        master_weight = unpack_xdr_array(
            operation_xdr_object.body.setOptionsOp.masterWeight
        )
        low_threshold = unpack_xdr_array(
            operation_xdr_object.body.setOptionsOp.lowThreshold
        )
        med_threshold = unpack_xdr_array(
            operation_xdr_object.body.setOptionsOp.medThreshold
        )
        high_threshold = unpack_xdr_array(
            operation_xdr_object.body.setOptionsOp.highThreshold
        )
        home_domain = unpack_xdr_array(
            operation_xdr_object.body.setOptionsOp.homeDomain
        )

        if home_domain:
            home_domain = home_domain.decode("utf-8")

        signer = None
        signer_xdr_object = operation_xdr_object.body.setOptionsOp.signer
        if signer_xdr_object:
            signer = Signer.from_xdr_object(signer_xdr_object[0])

        op = cls(
            inflation_dest=inflation_dest,
            clear_flags=clear_flags,
            set_flags=set_flags,
            master_weight=master_weight,
            low_threshold=low_threshold,
            med_threshold=med_threshold,
            high_threshold=high_threshold,
            home_domain=home_domain,
            signer=signer,
            source=source,
        )
        op._source_muxed = Operation.get_source_muxed_from_xdr_obj(operation_xdr_object)
        return op



import time
import warnings
from decimal import Decimal
from typing import List, Union, Optional

from .signer_key import SignerKey
from .account import Account
from .asset import Asset
from .exceptions import ValueError
from .fee_bump_transaction import FeeBumpTransaction
from .fee_bump_transaction_envelope import FeeBumpTransactionEnvelope
from .keypair import Keypair
from .memo import *
from .network import Network
from .operation import *
from .price import Price
from .signer import Signer
from .time_bounds import TimeBounds
from .transaction import Transaction
from .transaction_envelope import TransactionEnvelope
from .utils import hex_to_bytes
from .xdr import Xdr

__all__ = ["TransactionBuilder"]


class TransactionBuilder:
    """Transaction builder helps constructs a new :class:`TransactionEnvelope
    <stellar_sdk.transaction_envelope.TransactionEnvelope>` using the given
    :class:`Account <stellar_sdk.account.Account>` as the transaction's "source account". The transaction will use
    the current sequence number of the given account as its sequence number and increment the given account's
    sequence number by one. The given source account must include a private key for signing the transaction or
    an error will be thrown.

    Be careful about **unsubmitted transactions**! When you build a transaction, stellar-sdk automatically
    increments the source account's sequence number. If you end up not submitting this transaction and submitting
    another one instead, it'll fail due to the sequence number being wrong. So if you decide not to use a built
    transaction, make sure to update the source account's sequence number with :meth:`stellar_sdk.Server.load_account`
    before creating another transaction.

    :param source_account: The source account for this transaction.
    :param network_passphrase: The network to connect to for verifying and retrieving additional attributes from.
        Defaults to **Test SDF Network ; September 2015**.
    :param base_fee: Base fee in stroops. The network base fee is obtained by default from the latest ledger.
        Transaction fee is equal to base fee times number of operations in this transaction.
    :param v1: When this value is set to True, V1 transactions will be generated,
        otherwise V0 transactions will be generated.
        See `CAP-0015 <https://github.com/stellar/stellar-protocol/blob/master/core/cap-0015.md>`_ for more information.
    """

    # TODO: add an example
    def __init__(
        self,
        source_account: Account,
        network_passphrase: str = Network.TESTNET_NETWORK_PASSPHRASE,
        base_fee: int = 100,
        v1: bool = True,
    ):
        self.source_account: Account = source_account
        self.base_fee: int = base_fee
        self.network_passphrase: str = network_passphrase
        self.operations: List[Operation] = []
        self.time_bounds: Optional[TimeBounds] = None
        self.memo: Memo = NoneMemo()
        self.v1: bool = v1

    def build(self) -> TransactionEnvelope:
        """This will build the transaction envelope.
        It will also increment the source account's sequence number by 1.

        :return: The transaction envelope.
        """
        source = self.source_account.account_id
        sequence = self.source_account.sequence + 1
        transaction = Transaction(
            source=source,
            sequence=sequence,
            fee=self.base_fee * len(self.operations),
            operations=self.operations,
            memo=self.memo,
            time_bounds=self.time_bounds,
            v1=self.v1,
        )
        transaction_envelope = TransactionEnvelope(
            transaction=transaction, network_passphrase=self.network_passphrase
        )
        self.source_account.increment_sequence_number()
        return transaction_envelope

    @staticmethod
    def build_fee_bump_transaction(
        fee_source: Union[Keypair, str],
        base_fee: int,
        inner_transaction_envelope: TransactionEnvelope,
        network_passphrase: str = Network.TESTNET_NETWORK_PASSPHRASE,
    ) -> "FeeBumpTransactionEnvelope":
        """Create a
        :py:class:`FeeBumpTransactionEnvelope <stellar_sdk.fee_bump_transaction_envelope.FeeBumpTransactionEnvelope>`
        object.

        See `CAP-0015 <https://github.com/stellar/stellar-protocol/blob/master/core/cap-0015.md>`_ for more information.

        :param fee_source: The account paying for the transaction.
        :param base_fee: The max fee willing to pay per operation in inner transaction (**in stroops**).
        :param inner_transaction_envelope: The TransactionEnvelope to be bumped by the fee bump transaction.
        :param network_passphrase: The network to connect to for verifying and retrieving additional attributes from.
        :return: a :class:`TransactionBuilder <stellar_sdk.transaction_envelope.TransactionBuilder>` via the XDR object.
        """
        fee_bump_transaction = FeeBumpTransaction(
            fee_source=fee_source,
            base_fee=base_fee,
            inner_transaction_envelope=inner_transaction_envelope,
        )
        transaction_envelope = FeeBumpTransactionEnvelope(
            transaction=fee_bump_transaction, network_passphrase=network_passphrase,
        )
        return transaction_envelope

    @staticmethod
    def from_xdr(
        xdr: str, network_passphrase: str
    ) -> Union["TransactionBuilder", FeeBumpTransactionEnvelope]:
        """Create a :class:`TransactionBuilder
        <stellar_sdk.transaction_envelope.TransactionBuilder>` or
        :py:class:`FeeBumpTransactionEnvelope <stellar_sdk.fee_bump_transaction_envelope.FeeBumpTransactionEnvelope>`
        via an XDR object.

        In addition, if xdr is not of
        :py:class:`TransactionEnvelope <stellar_sdk.transaction_envelope.TransactionEnvelope>`,
        it sets the fields of this builder (the transaction envelope,
        transaction, operations, source, etc.) to all of the fields in the
        provided XDR transaction envelope, but the signature will not be added to it.

        :param xdr: The XDR object representing the transaction envelope to
            which this builder is setting its state to.
        :param network_passphrase: The network to connect to for verifying and retrieving additional attributes from.
        :return: a :class:`TransactionBuilder <stellar_sdk.transaction_envelope.TransactionBuilder>` or
            :py:class:`FeeBumpTransactionEnvelope <stellar_sdk.fee_bump_transaction_envelope.FeeBumpTransactionEnvelope>`
            via the XDR object.
        """
        xdr_object = Xdr.types.TransactionEnvelope.from_xdr(xdr)
        te_type = xdr_object.type
        if te_type == Xdr.const.ENVELOPE_TYPE_TX_FEE_BUMP:
            return FeeBumpTransactionEnvelope.from_xdr_object(
                xdr_object, network_passphrase
            )
        transaction_envelope = TransactionEnvelope.from_xdr(
            xdr=xdr, network_passphrase=network_passphrase
        )

        source_account = Account(
            transaction_envelope.transaction.source.public_key,
            transaction_envelope.transaction.sequence - 1,
        )
        transaction_builder = TransactionBuilder(
            source_account=source_account,
            network_passphrase=network_passphrase,
            base_fee=int(
                transaction_envelope.transaction.fee
                / len(transaction_envelope.transaction.operations)
            ),
            v1=transaction_envelope.transaction.v1,
        )
        transaction_builder.time_bounds = transaction_envelope.transaction.time_bounds
        transaction_builder.operations = transaction_envelope.transaction.operations
        transaction_builder.memo = transaction_envelope.transaction.memo
        return transaction_builder

    def add_time_bounds(self, min_time: int, max_time: int) -> "TransactionBuilder":
        """Add a time bound to this transaction.

        Add a UNIX timestamp, determined by ledger time, of a lower and
        upper bound of when this transaction will be valid. If a transaction is
        submitted too early or too late, it will fail to make it into the
        transaction set. maxTime equal 0 means that it's not set.

        :param min_time: the UNIX timestamp (in seconds)
        :param max_time: the UNIX timestamp (in seconds)
        :return: This builder instance.

        """
        self.time_bounds = TimeBounds(min_time, max_time)
        return self

    def set_timeout(self, timeout: int) -> "TransactionBuilder":
        """Set timeout for the transaction, actually set a TimeBounds.

        :param timeout: timeout in second.
        :return: This builder instance.
        :raises:
            :exc:`ValueError <stellar_sdk.exceptions.ValueError>`: if time_bound is already set.
        """
        if self.time_bounds:
            raise ValueError(
                "TimeBounds has been already set - setting timeout would overwrite it."
            )
        timeout_timestamp = int(time.time()) + timeout
        self.time_bounds = TimeBounds(min_time=0, max_time=timeout_timestamp)
        return self

    def add_memo(self, memo: Memo) -> "TransactionBuilder":
        """Set the memo for the transaction build by this :class:`Builder`.

        :param memo: A memo to add to this transaction.
        :return: This builder instance.

        """
        self.memo = memo
        return self

    def add_text_memo(self, memo_text: Union[str, bytes]) -> "TransactionBuilder":
        """Set the memo for the transaction to a new :class:`TextMemo
        <stellar_sdk.memo.TextMemo>`.

        :param memo_text: The text for the memo to add.
        :return: This builder instance.
        :raises: :exc:`MemoInvalidException <stellar_sdk.exceptions.MemoInvalidException>`:
            if ``memo_text`` is not a valid text memo.
        """
        memo = TextMemo(memo_text)
        return self.add_memo(memo)

    def add_id_memo(self, memo_id: int) -> "TransactionBuilder":
        """Set the memo for the transaction to a new :class:`IdMemo
        <stellar_sdk.memo.IdMemo>`.

        :param memo_id: A 64 bit unsigned integer to set as the memo.
        :return: This builder instance.
        :raises:
            :exc:`MemoInvalidException <stellar_sdk.exceptions.MemoInvalidException>`:
            if ``memo_id`` is not a valid id memo.

        """
        memo = IdMemo(memo_id)
        return self.add_memo(memo)

    def add_hash_memo(self, memo_hash: Union[bytes, str]) -> "TransactionBuilder":
        """Set the memo for the transaction to a new :class:`HashMemo
        <stellar_sdk.memo.HashMemo>`.

        :param memo_hash: A 32 byte hash or hex encoded string to use as the memo.
        :return: This builder instance.
        :raises:
            :exc:`MemoInvalidException <stellar_sdk.exceptions.MemoInvalidException>`:
            if ``memo_hash`` is not a valid hash memo.
        """
        memo = HashMemo(hex_to_bytes(memo_hash))
        return self.add_memo(memo)

    def add_return_hash_memo(
        self, memo_return: Union[bytes, str]
    ) -> "TransactionBuilder":
        """Set the memo for the transaction to a new :class:`RetHashMemo
        <stellar_sdk.memo.ReturnHashMemo>`.

        :param memo_return: A 32 byte hash or hex encoded string intended to be interpreted as
            the hash of the transaction the sender is refunding.
        :return: This builder instance.
        :raises:
            :exc:`MemoInvalidException <stellar_sdk.exceptions.MemoInvalidException>`:
            if ``memo_return`` is not a valid return hash memo.
        """
        memo = ReturnHashMemo(hex_to_bytes(memo_return))
        return self.add_memo(memo)

    def append_operation(self, operation: Operation) -> "TransactionBuilder":
        """Add an operation to the builder instance

        :param operation: an operation
        :return: This builder instance.
        """
        # TODO: LOG HERE
        self.operations.append(operation)
        return self

    def append_create_account_op(
        self,
        destination: str,
        starting_balance: Union[str, Decimal],
        source: str = None,
    ) -> "TransactionBuilder":
        """Append a :class:`CreateAccount
        <stellar_sdk.operation.CreateAccount>` operation to the list of
        operations.

        :param destination: Account address that is created and funded.
        :param starting_balance: Amount of XLM to send to the newly created
            account. This XLM comes from the source account.
        :param source: The source address to deduct funds from to fund the
            new account.
        :return: This builder instance.

        """
        op = CreateAccount(destination, starting_balance, source)
        return self.append_operation(op)

    def append_change_trust_op(
        self,
        asset_code: str,
        asset_issuer: str,
        limit: Union[str, Decimal] = None,
        source: str = None,
    ) -> "TransactionBuilder":
        """Append a :class:`ChangeTrust <stellar_sdk.operation.ChangeTrust>`
        operation to the list of operations.

        :param asset_issuer: The issuer address for the asset.
        :param asset_code: The asset code for the asset.
        :param limit: The limit of the new trustline.
        :param source: The source address to add the trustline to.
        :return: This builder instance.

        """
        asset = Asset(asset_code, asset_issuer)
        op = ChangeTrust(asset, limit, source)
        return self.append_operation(op)

    def append_payment_op(
        self,
        destination: str,
        amount: Union[str, Decimal],
        asset_code: str = "XLM",
        asset_issuer: Optional[str] = None,
        source: str = None,
    ) -> "TransactionBuilder":
        """Append a :class:`Payment <stellar_sdk.operation.Payment>` operation
        to the list of operations.

        :param destination: Account address that receives the payment.
        :param amount: The amount of the currency to send in the payment.
        :param asset_code: The asset code for the asset to send.
        :param asset_issuer: The address of the issuer of the asset.
        :param source: The source address of the payment.
        :return: This builder instance.

        """
        asset = Asset(code=asset_code, issuer=asset_issuer)
        op = Payment(destination, asset, amount, source)
        return self.append_operation(op)

    def append_path_payment_op(
        self,
        destination: str,
        send_code: str,
        send_issuer: Optional[str],
        send_max: Union[str, Decimal],
        dest_code: str,
        dest_issuer: Optional[str],
        dest_amount: Union[str, Decimal],
        path: List[Asset],
        source: str = None,
    ) -> "TransactionBuilder":
        """Append a :class:`PathPayment <stellar_sdk.operation.PathPayment>`
        operation to the list of operations.

        :param destination: The destination address (Account ID) for the
            payment.
        :param send_code: The asset code for the source asset deducted from
            the source account.
        :param send_issuer: The address of the issuer of the source asset.
        :param send_max: The maximum amount of send asset to deduct
            (excluding fees).
        :param dest_code: The asset code for the final destination asset
            sent to the recipient.
        :param dest_issuer: Account address that receives the payment.
        :param dest_amount: The amount of destination asset the destination
            account receives.
        :param path: A list of Asset objects to use as the path.
        :param source: The source address of the path payment.
        :return: This builder instance.

        """

        warnings.warn(
            "Will be removed in version v3.0.0, "
            "use stellar_sdk.transaction_builder.append_path_payment_strict_receive_op",
            DeprecationWarning,
        )

        send_asset = Asset(send_code, send_issuer)
        dest_asset = Asset(dest_code, dest_issuer)

        assets = []
        for asset in path:
            assets.append(asset)
        op = PathPayment(
            destination=destination,
            send_asset=send_asset,
            send_max=send_max,
            dest_asset=dest_asset,
            dest_amount=dest_amount,
            path=assets,
            source=source,
        )
        return self.append_operation(op)

    def append_path_payment_strict_receive_op(
        self,
        destination: str,
        send_code: str,
        send_issuer: Optional[str],
        send_max: Union[str, Decimal],
        dest_code: str,
        dest_issuer: Optional[str],
        dest_amount: Union[str, Decimal],
        path: List[Asset],
        source: str = None,
    ) -> "TransactionBuilder":
        """Append a :class:`PathPaymentStrictReceive <stellar_sdk.operation.PathPaymentStrictReceive>`
        operation to the list of operations.

        :param destination: The destination address (Account ID) for the
            payment.
        :param send_code: The asset code for the source asset deducted from
            the source account.
        :param send_issuer: The address of the issuer of the source asset.
        :param send_max: The maximum amount of send asset to deduct
            (excluding fees).
        :param dest_code: The asset code for the final destination asset
            sent to the recipient.
        :param dest_issuer: Account address that receives the payment.
        :param dest_amount: The amount of destination asset the destination
            account receives.
        :param path: A list of Asset objects to use as the path.
        :param source: The source address of the path payment.
        :return: This builder instance.

        """

        send_asset = Asset(send_code, send_issuer)
        dest_asset = Asset(dest_code, dest_issuer)

        assets = []
        for asset in path:
            assets.append(asset)
        op = PathPaymentStrictReceive(
            destination=destination,
            send_asset=send_asset,
            send_max=send_max,
            dest_asset=dest_asset,
            dest_amount=dest_amount,
            path=assets,
            source=source,
        )
        return self.append_operation(op)

    def append_path_payment_strict_send_op(
        self,
        destination: str,
        send_code: str,
        send_issuer: Optional[str],
        send_amount: Union[str, Decimal],
        dest_code: str,
        dest_issuer: Optional[str],
        dest_min: Union[str, Decimal],
        path: List[Asset],
        source: str = None,
    ) -> "TransactionBuilder":
        """Append a :class:`PathPaymentStrictSend <stellar_sdk.operation.PathPaymentStrictSend>`
        operation to the list of operations.

        :param destination: The destination address (Account ID) for the
            payment.
        :param send_code: The asset code for the source asset deducted from
            the source account.
        :param send_issuer: The address of the issuer of the source asset.
        :param send_amount: Amount of send_asset to send.
        :param dest_code: The asset code for the final destination asset
            sent to the recipient.
        :param dest_issuer: Account address that receives the payment.
        :param dest_min: The minimum amount of dest_asset to be received.
        :param path: A list of Asset objects to use as the path.
        :param source: The source address of the path payment.
        :return: This builder instance.

        """

        send_asset = Asset(send_code, send_issuer)
        dest_asset = Asset(dest_code, dest_issuer)

        assets = []
        for asset in path:
            assets.append(asset)
        op = PathPaymentStrictSend(
            destination=destination,
            send_asset=send_asset,
            send_amount=send_amount,
            dest_asset=dest_asset,
            dest_min=dest_min,
            path=assets,
            source=source,
        )
        return self.append_operation(op)

    def append_allow_trust_op(
        self,
        trustor: str,
        asset_code: str,
        authorize: Union[TrustLineEntryFlag, bool],
        source: str = None,
    ) -> "TransactionBuilder":
        """Append an :class:`AllowTrust <stellar_sdk.operation.AllowTrust>`
        operation to the list of operations.

        :param trustor: The account of the recipient of the trustline.
        :param asset_code:  The asset of the trustline the source account
            is authorizing. For example, if an anchor wants to allow another
            account to hold its USD credit, the type is USD:anchor.
        :param authorize: `True` to authorize the line, `False` to deauthorize，if you need further control,
            you can also use :class:`stellar_sdk.operation.allow_trust.TrustLineEntryFlag`.
        :param source: The source address that is establishing the trust in
            the allow trust operation.
        :return: This builder instance.

        """
        op = AllowTrust(trustor, asset_code, authorize, source)
        return self.append_operation(op)

    def append_set_options_op(
        self,
        inflation_dest: str = None,
        clear_flags: Union[int, Flag] = None,
        set_flags: Union[int, Flag] = None,
        master_weight: int = None,
        low_threshold: int = None,
        med_threshold: int = None,
        high_threshold: int = None,
        home_domain: str = None,
        signer: Signer = None,
        source: str = None,
    ) -> "TransactionBuilder":
        """Append a :class:`SetOptions <stellar_sdk.operation.SetOptions>`
        operation to the list of operations.

        :param inflation_dest: Account of the inflation destination.
        :param clear_flags: Indicates which flags to clear. For details about the flags,
            please refer to the `accounts doc <https://www.stellar.org/developers/guides/concepts/accounts.html>`__.
            The `bit mask <https://en.wikipedia.org/wiki/Bit_field>`_ integer subtracts from the existing flags of the account.
            This allows for setting specific bits without knowledge of existing flags, you can also use
            :class:`stellar_sdk.operation.set_options.Flag`
            - AUTHORIZATION_REQUIRED = 1
            - AUTHORIZATION_REVOCABLE = 2
            - AUTHORIZATION_IMMUTABLE = 4
        :param set_flags: Indicates which flags to set. For details about the flags,
            please refer to the `accounts doc <https://www.stellar.org/developers/guides/concepts/accounts.html>`__.
            The bit mask integer adds onto the existing flags of the account.
            This allows for setting specific bits without knowledge of existing flags, you can also use
            :class:`stellar_sdk.operation.set_options.Flag`
            - AUTHORIZATION_REQUIRED = 1
            - AUTHORIZATION_REVOCABLE = 2
            - AUTHORIZATION_IMMUTABLE = 4
        :param master_weight: A number from 0-255 (inclusive) representing the weight of the master key.
            If the weight of the master key is updated to 0, it is effectively disabled.
        :param low_threshold: A number from 0-255 (inclusive) representing the threshold this account sets on all
            operations it performs that have `a low threshold <https://www.stellar.org/developers/guides/concepts/multi-sig.html>`_.
        :param med_threshold: A number from 0-255 (inclusive) representing the threshold this account sets on all
            operations it performs that have `a medium threshold <https://www.stellar.org/developers/guides/concepts/multi-sig.html>`_.
        :param high_threshold: A number from 0-255 (inclusive) representing the threshold this account sets on all
            operations it performs that have `a high threshold <https://www.stellar.org/developers/guides/concepts/multi-sig.html>`_.
        :param home_domain: sets the home domain used for
            reverse `federation <https://www.stellar.org/developers/guides/concepts/federation.html>`_ lookup.
        :param signer: Add, update, or remove a signer from the account.
        :param source: The source account (defaults to transaction source).
        :return: This builder instance.

        """
        op = SetOptions(
            inflation_dest=inflation_dest,
            clear_flags=clear_flags,
            set_flags=set_flags,
            master_weight=master_weight,
            low_threshold=low_threshold,
            med_threshold=med_threshold,
            high_threshold=high_threshold,
            signer=signer,
            home_domain=home_domain,
            source=source,
        )
        return self.append_operation(op)

    def append_ed25519_public_key_signer(
        self, account_id: str, weight: int, source: str = None
    ) -> "TransactionBuilder":
        """Add a ed25519 public key signer to an account.

        Add a ed25519 public key signer to an account via a :class:`SetOptions
        <stellar_sdk.operation.SetOptions` operation. This is a helper
        function for :meth:`append_set_options_op`.

        :param account_id: The account id of the new ed25519_public_key signer.
        :param weight: The weight of the new signer.
        :param source: The source account that is adding a signer to its
            list of signers.
        :return: This builder instance.

        """
        signer = Signer.ed25519_public_key(account_id=account_id, weight=weight)
        return self.append_set_options_op(signer=signer, source=source)

    def append_hashx_signer(
        self, sha256_hash: Union[bytes, str], weight: int, source: str = None,
    ) -> "TransactionBuilder":
        """Add a sha256 hash(HashX) signer to an account.

        Add a HashX signer to an account via a :class:`SetOptions
        <stellar_sdk.operation.SetOptions` operation. This is a helper
        function for :meth:`append_set_options_op`.

        :param sha256_hash: The address of the new sha256 hash(hashX) signer,
            a 32 byte hash or hex encoded string.
        :param weight: The weight of the new signer.
        :param source: The source account that is adding a signer to its
            list of signers.
        :return: This builder instance.

        """
        signer = Signer.sha256_hash(
            sha256_hash=hex_to_bytes(sha256_hash), weight=weight
        )
        return self.append_set_options_op(signer=signer, source=source)

    def append_pre_auth_tx_signer(
        self, pre_auth_tx_hash: Union[str, bytes], weight: int, source=None
    ) -> "TransactionBuilder":
        """Add a PreAuthTx signer to an account.

        Add a PreAuthTx signer to an account via a :class:`SetOptions
        <stellar_sdk.operation.SetOptions` operation. This is a helper
        function for :meth:`append_set_options_op`.

        :param pre_auth_tx_hash: The address of the new preAuthTx signer - obtained by calling ``hash`` on
            the :class:`TransactionEnvelope <stellar_sdk.transaction_envelope.TransactionEnvelope>`,
            a 32 byte hash or hex encoded string.
        :param weight: The weight of the new signer.
        :param source: The source account that is adding a signer to its
            list of signers.
        :return: This builder instance.

        """
        signer = Signer.pre_auth_tx(
            pre_auth_tx_hash=hex_to_bytes(pre_auth_tx_hash), weight=weight
        )
        return self.append_set_options_op(signer=signer, source=source)

    def append_manage_buy_offer_op(
        self,
        selling_code: str,
        selling_issuer: Optional[str],
        buying_code: str,
        buying_issuer: Optional[str],
        amount: Union[str, Decimal],
        price: Union[str, Decimal, Price],
        offer_id: int = 0,
        source: str = None,
    ) -> "TransactionBuilder":
        """Append a :class:`ManageBuyOffer <stellar_sdk.operation.ManageBuyOffer>`
        operation to the list of operations.

        :param selling_code: The asset code for the asset the offer creator
            is selling.
        :param selling_issuer: The issuing address for the asset the offer
            creator is selling.
        :param buying_code: The asset code for the asset the offer creator
            is buying.
        :param buying_issuer: The issuing address for the asset the offer
            creator is buying.
        :param amount: Amount being bought. if set to. Set to 0 if you want
            to delete an existing offer.
        :param price: Price of thing being bought in terms of what you are selling.
        :param offer_id: The ID of the offer. 0 for new offer. Set to
            existing offer ID to update or delete.
        :param source: The source address that is managing a buying offer on
            Stellar's distributed exchange.
        :return: This builder instance.

        """
        selling = Asset(selling_code, selling_issuer)
        buying = Asset(buying_code, buying_issuer)
        op = ManageBuyOffer(
            selling=selling,
            buying=buying,
            amount=amount,
            price=price,
            offer_id=offer_id,
            source=source,
        )
        return self.append_operation(op)

    def append_manage_sell_offer_op(
        self,
        selling_code: str,
        selling_issuer: Optional[str],
        buying_code: str,
        buying_issuer: Optional[str],
        amount: Union[str, Decimal],
        price: Union[str, Price, Decimal],
        offer_id: int = 0,
        source: str = None,
    ) -> "TransactionBuilder":
        """Append a :class:`ManageSellOffer <stellar_sdk.operation.ManageSellOffer>`
        operation to the list of operations.

        :param selling_code: The asset code for the asset the offer creator
            is selling.
        :param selling_issuer: The issuing address for the asset the offer
            creator is selling.
        :param buying_code: The asset code for the asset the offer creator
            is buying.
        :param buying_issuer: The issuing address for the asset the offer
            creator is buying.
        :param amount: Amount of the asset being sold. Set to 0 if you want
            to delete an existing offer.
        :param price: Price of 1 unit of selling in terms of buying.
        :param offer_id: The ID of the offer. 0 for new offer. Set to
            existing offer ID to update or delete.
        :param source: The source address that is managing an offer on
            Stellar's distributed exchange.
        :return: This builder instance.

        """
        selling = Asset(selling_code, selling_issuer)
        buying = Asset(buying_code, buying_issuer)
        op = ManageSellOffer(
            selling=selling,
            buying=buying,
            amount=amount,
            price=price,
            offer_id=offer_id,
            source=source,
        )
        return self.append_operation(op)

    def append_create_passive_sell_offer_op(
        self,
        selling_code: str,
        selling_issuer: Optional[str],
        buying_code: str,
        buying_issuer: Optional[str],
        amount: Union[str, Decimal],
        price: Union[str, Price, Decimal],
        source: str = None,
    ) -> "TransactionBuilder":
        """Append a :class:`CreatePassiveSellOffer
        <stellar_sdk.operation.CreatePassiveSellOffer>` operation to the list of
        operations.

        :param selling_code: The asset code for the asset the offer creator
            is selling.
        :param selling_issuer: The issuing address for the asset the offer
            creator is selling.
        :param buying_code: The asset code for the asset the offer creator
            is buying.
        :param buying_issuer: The issuing address for the asset the offer
            creator is buying.
        :param amount: Amount of the asset being sold. Set to 0 if you want
            to delete an existing offer.
        :param price: Price of 1 unit of selling in terms of buying.
        :param source: The source address that is creating a passive offer
            on Stellar's distributed exchange.
        :return: This builder instance.

        """

        selling = Asset(selling_code, selling_issuer)
        buying = Asset(buying_code, buying_issuer)
        op = CreatePassiveSellOffer(selling, buying, amount, price, source)
        return self.append_operation(op)

    def append_account_merge_op(
        self, destination: str, source: str = None,
    ) -> "TransactionBuilder":
        """Append a :class:`AccountMerge
        <stellar_sdk.operation.AccountMerge>` operation to the list of
        operations.

        :param destination: The ID of the offer. 0 for new offer. Set to
            existing offer ID to update or delete.
        :param source: The source address that is being merged into the
            destination account.
        :return: This builder instance.

        """
        op = AccountMerge(destination, source)
        return self.append_operation(op)

    def append_inflation_op(self, source: str = None) -> "TransactionBuilder":
        """Append a :class:`Inflation
        <stellar_sdk.operation.Inflation>` operation to the list of
        operations.

        :param source: The source address that is running the inflation
            operation.
        :return: This builder instance.

        """
        op = Inflation(source)
        return self.append_operation(op)

    def append_manage_data_op(
        self, data_name: str, data_value: Union[str, bytes, None], source: str = None,
    ) -> "TransactionBuilder":
        """Append a :class:`ManageData <stellar_sdk.operation.ManageData>`
        operation to the list of operations.

        :param data_name: String up to 64 bytes long. If this is a new Name
            it will add the given name/value pair to the account. If this Name
            is already present then the associated value will be modified.
        :param data_value: If not present then the existing
            Name will be deleted. If present then this value will be set in the
            DataEntry. Up to 64 bytes long.
        :param source: The source account on which data is being managed.
            operation.
        :return: This builder instance.

        """
        op = ManageData(data_name, data_value, source)
        return self.append_operation(op)

    def append_bump_sequence_op(
        self, bump_to: int, source: str = None
    ) -> "TransactionBuilder":
        """Append a :class:`BumpSequence <stellar_sdk.operation.BumpSequence>`
        operation to the list of operations.

        :param bump_to: Sequence number to bump to.
        :param source: The source address that is running the inflation operation.
        :return: This builder instance.

        """
        op = BumpSequence(bump_to, source)
        return self.append_operation(op)

    def append_create_claimable_balance_op(
        self,
        asset: Asset,
        amount: Union[str, Decimal],
        claimants: List[Claimant],
        source: str = None,
    ) -> "TransactionBuilder":
        """Append a :class:`CreateClaimableBalance <stellar_sdk.operation.CreateClaimableBalance>`
        operation to the list of operations.

        :param asset: The asset for the claimable balance.
        :param amount: the amount of the asset.
        :param claimants: A list of Claimants.
        :param source: The source account (defaults to transaction source).
        """
        op = CreateClaimableBalance(asset, amount, claimants, source)
        return self.append_operation(op)

    def append_claim_claimable_balance_op(
        self, balance_id: str, source: str = None
    ) -> "TransactionBuilder":
        """Append a :class:`ClaimClaimableBalance <stellar_sdk.operation.ClaimClaimableBalance>`
        operation to the list of operations.

        :param balance_id: The claimable balance id to be claimed.
        :param source: The source account (defaults to transaction source).
        """
        op = ClaimClaimableBalance(balance_id, source)
        return self.append_operation(op)

    def append_begin_sponsoring_future_reserves_op(
        self, sponsored_id: str, source: str = None
    ) -> "TransactionBuilder":
        """Append a :class:`BeginSponsoringFutureReserves <stellar_sdk.operation.BeginSponsoringFutureReserves>`
        operation to the list of operations.

        :param sponsored_id: The sponsored account id.
        :param source: The source account (defaults to transaction source).
        :return: This builder instance.
        """
        op = BeginSponsoringFutureReserves(sponsored_id, source)
        return self.append_operation(op)

    def append_end_sponsoring_future_reserves_op(
        self, source: str = None
    ) -> "TransactionBuilder":
        """Append a :class:`EndSponsoringFutureReserves <stellar_sdk.operation.EndSponsoringFutureReserves>`
        operation to the list of operations.

        :param source: The source account (defaults to transaction source).
        :return: This builder instance.
        """
        op = EndSponsoringFutureReserves(source)
        return self.append_operation(op)

    def append_revoke_account_sponsorship_op(
        self, account_id: str, source: str = None
    ) -> "TransactionBuilder":
        """Append a :class:`EndSponsoringFutureReserves <stellar_sdk.operation.EndSponsoringFutureReserves>` operation
        for an account to the list of operations.

        :param account_id: The sponsored account ID.
        :param source: The source account (defaults to transaction source).
        :return: This builder instance.
        """
        op = RevokeSponsorship.revoke_account_sponsorship(account_id, source)
        return self.append_operation(op)

    def append_revoke_trustline_sponsorship_op(
        self, account_id: str, asset: Asset, source: str = None
    ) -> "TransactionBuilder":
        """Append a :class:`EndSponsoringFutureReserves <stellar_sdk.operation.EndSponsoringFutureReserves>` operation
        for a trustline to the list of operations.

        :param account_id: The account ID which owns the trustline.
        :param asset: The asset in the trustline.
        :param source: The source account (defaults to transaction source).
        :return: This builder instance.
        """
        op = RevokeSponsorship.revoke_trustline_sponsorship(account_id, asset, source)
        return self.append_operation(op)

    def append_revoke_offer_sponsorship_op(
        self, seller_id: str, offer_id: int, source: str = None
    ) -> "TransactionBuilder":
        """Append a :class:`EndSponsoringFutureReserves <stellar_sdk.operation.EndSponsoringFutureReserves>` operation
        for an offer to the list of operations.

        :param seller_id: The account ID which created the offer.
        :param offer_id: The offer ID.
        :param source: The source account (defaults to transaction source).
        :return: This builder instance.
        """
        op = RevokeSponsorship.revoke_offer_sponsorship(seller_id, offer_id, source)
        return self.append_operation(op)

    def append_revoke_data_sponsorship_op(
        self, account_id: str, data_name: str, source: str = None
    ) -> "TransactionBuilder":
        """Append a :class:`EndSponsoringFutureReserves <stellar_sdk.operation.EndSponsoringFutureReserves>` operation
        for a data entry to the list of operations.

        :param account_id: The account ID which owns the data entry.
        :param data_name: The name of the data entry
        :param source: The source account (defaults to transaction source).
        :return: This builder instance.
        """
        op = RevokeSponsorship.revoke_data_sponsorship(account_id, data_name, source)
        return self.append_operation(op)

    def append_revoke_claimable_balance_sponsorship_op(
        self, claimable_balance_id: str, source: str = None
    ) -> "TransactionBuilder":
        """Append a :class:`EndSponsoringFutureReserves <stellar_sdk.operation.EndSponsoringFutureReserves>` operation
        for a claimable to the list of operations.

        :param claimable_balance_id: The sponsored claimable balance ID.
        :param source: The source account (defaults to transaction source).
        :return: This builder instance.
        """
        op = RevokeSponsorship.revoke_claimable_balance_sponsorship(
            claimable_balance_id, source
        )
        return self.append_operation(op)

    def append_revoke_ed25519_public_key_signer_sponsorship_op(
        self, account_id: str, signer_key: str, source: str = None
    ) -> "TransactionBuilder":
        """Append a :class:`EndSponsoringFutureReserves <stellar_sdk.operation.EndSponsoringFutureReserves>` operation
        for a ed25519_public_key signer to the list of operations.

        :param account_id: The account ID where the signer sponsorship is being removed from.
        :param signer_key: The account id of the ed25519_public_key signer.
        :param source: The source account (defaults to transaction source).
        :return: This builder instance.
        """
        signer_key = SignerKey.ed25519_public_key(signer_key)
        op = RevokeSponsorship.revoke_signer_sponsorship(account_id, signer_key, source)
        return self.append_operation(op)

    def append_revoke_hashx_signer_sponsorship_op(
        self, account_id: str, signer_key: Union[bytes, str], source: str = None
    ) -> "TransactionBuilder":
        """Append a :class:`EndSponsoringFutureReserves <stellar_sdk.operation.EndSponsoringFutureReserves>` operation
        for a hashx signer to the list of operations.

        :param account_id: The account ID where the signer sponsorship is being removed from.
        :param signer_key: The account id of the hashx signer.
        :param source: The source account (defaults to transaction source).
        :return: This builder instance.
        """
        signer_key = SignerKey.sha256_hash(hex_to_bytes(signer_key))
        op = RevokeSponsorship.revoke_signer_sponsorship(account_id, signer_key, source)
        return self.append_operation(op)

    def append_revoke_pre_auth_tx_signer_sponsorship_op(
        self, account_id: str, signer_key: Union[bytes, str], source: str = None
    ) -> "TransactionBuilder":
        """Append a :class:`EndSponsoringFutureReserves <stellar_sdk.operation.EndSponsoringFutureReserves>` operation
        for a pre_auth_tx signer to the list of operations.

        :param account_id: The account ID where the signer sponsorship is being removed from.
        :param signer_key: The account id of the pre_auth_tx signer.
        :param source: The source account (defaults to transaction source).
        :return: This builder instance.
        """
        signer_key = SignerKey.pre_auth_tx(hex_to_bytes(signer_key))
        op = RevokeSponsorship.revoke_signer_sponsorship(account_id, signer_key, source)
        return self.append_operation(op)



import base64
import binascii
from enum import IntEnum
from typing import Optional

from .operation import Operation
from ..asset import Asset
from ..keypair import Keypair
from ..signer_key import SignerKey
from ..strkey import StrKey
from ..xdr import Xdr
from ..exceptions import ValueError
from .utils import check_ed25519_public_key
from ..xdr.StellarXDR_type import ClaimableBalanceID


class RevokeSponsorshipType(IntEnum):
    """Currently supported RevokeSponsorship types.
    """

    ACCOUNT = 0
    TRUSTLINE = 1
    OFFER = 2
    DATA = 3
    CLAIMABLE_BALANCE = 4
    SIGNER = 5


class TrustLine:
    def __init__(self, account_id: str, asset: Asset) -> None:
        check_ed25519_public_key(account_id)
        self.account_id = account_id
        self.asset = asset


class Offer:
    def __init__(self, seller_id: str, offer_id: int) -> None:
        check_ed25519_public_key(seller_id)
        self.seller_id = seller_id
        self.offer_id = offer_id


class Data:
    def __init__(self, account_id: str, data_name: str) -> None:
        check_ed25519_public_key(account_id)
        self.account_id = account_id
        self.data_name = data_name


class Signer:
    def __init__(self, account_id: str, signer_key: SignerKey) -> None:
        check_ed25519_public_key(account_id)
        self.account_id = account_id
        self.signer_key = signer_key


class RevokeSponsorship(Operation):
    """The :class:`RevokeSponsorship` object, which represents a RevokeSponsorship
    operation on Stellar's network.

    The logic of this operation depends on the state of the source account.

    If the source account is not sponsored or is sponsored by the owner of the specified entry or sub-entry,
    then attempt to revoke the sponsorship. If the source account is sponsored, the next step depends on whether the
    entry is sponsored or not. If it is sponsored, attempt to transfer the sponsorship to the sponsor
    of the source account. If the entry is not sponsored, then establish the sponsorship.
    See `Sponsored Reserves <https://developers.stellar.org/docs/glossary/sponsored-reserves/>_` for more information.

    See `Revoke Sponsorship
    <https://developers.stellar.org/docs/start/list-of-operations/#revoke-sponsorship>_`.

    Threshold: Medium

    :param revoke_sponsorship_type: The sponsored account id.
    :param account_id: The sponsored account ID.
    :param trustline: The sponsored trustline.
    :param offer: The sponsored offer.
    :param data: The sponsored data.
    :param claimable_balance_id: The sponsored claimable balance.
    :param signer: The sponsored signer.
    :param source: The source account (defaults to transaction source).
    """

    def __init__(
        self,
        revoke_sponsorship_type: RevokeSponsorshipType,
        account_id: Optional[str],
        trustline: Optional[TrustLine],
        offer: Optional[Offer],
        data: Optional[Data],
        claimable_balance_id: Optional[str],
        signer: Optional[Signer],
        source: str = None,
    ) -> None:
        super().__init__(source)
        self.revoke_sponsorship_type = revoke_sponsorship_type
        self.account_id = account_id
        self.trustline = trustline
        self.offer = offer
        self.data = data
        self.claimable_balance_id = claimable_balance_id
        self.signer = signer

    @classmethod
    def type_code(cls) -> int:
        return Xdr.const.REVOKE_SPONSORSHIP

    @classmethod
    def revoke_account_sponsorship(cls, account_id: str, source: str = None):
        """Create a "revoke sponsorship" operation for an account.

        :param account_id: The sponsored account ID.
        :param source: The source account (defaults to transaction source).
        :return: A "revoke sponsorship" operation for an account.
        """
        return cls(
            revoke_sponsorship_type=RevokeSponsorshipType.ACCOUNT,
            account_id=account_id,
            trustline=None,
            offer=None,
            data=None,
            claimable_balance_id=None,
            signer=None,
            source=source,
        )

    @classmethod
    def revoke_trustline_sponsorship(
        cls, account_id: str, asset: Asset, source: str = None
    ):
        """Create a "revoke sponsorship" operation for a trustline.

        :param account_id: The account ID which owns the trustline.
        :param asset: The asset in the trustline.
        :param source: The source account (defaults to transaction source).
        :return: A "revoke sponsorship" operation for a trustline.
        """
        trustline = TrustLine(account_id=account_id, asset=asset)
        return cls(
            revoke_sponsorship_type=RevokeSponsorshipType.TRUSTLINE,
            account_id=None,
            trustline=trustline,
            offer=None,
            data=None,
            claimable_balance_id=None,
            signer=None,
            source=source,
        )

    @classmethod
    def revoke_offer_sponsorship(
        cls, seller_id: str, offer_id: int, source: str = None
    ):
        """Create a "revoke sponsorship" operation for an offer.

        :param seller_id: The account ID which created the offer.
        :param offer_id: The offer ID.
        :param source: The source account (defaults to transaction source).
        :return: A "revoke sponsorship" operation for an offer.
        """
        offer = Offer(seller_id=seller_id, offer_id=offer_id)
        return cls(
            revoke_sponsorship_type=RevokeSponsorshipType.OFFER,
            account_id=None,
            trustline=None,
            offer=offer,
            data=None,
            claimable_balance_id=None,
            signer=None,
            source=source,
        )

    @classmethod
    def revoke_data_sponsorship(
        cls, account_id: str, data_name: str, source: str = None
    ):
        """Create a "revoke sponsorship" operation for a data entry.

        :param account_id: The account ID which owns the data entry.
        :param data_name: The name of the data entry
        :param source: The source account (defaults to transaction source).
        :return: A "revoke sponsorship" operation for a data entry.
        """
        data = Data(account_id=account_id, data_name=data_name)
        return cls(
            revoke_sponsorship_type=RevokeSponsorshipType.DATA,
            account_id=None,
            trustline=None,
            offer=None,
            data=data,
            claimable_balance_id=None,
            signer=None,
            source=source,
        )

    @classmethod
    def revoke_claimable_balance_sponsorship(
        cls, claimable_balance_id: str, source: str = None
    ):
        """Create a "revoke sponsorship" operation for a claimable balance.

        :param claimable_balance_id: The sponsored claimable balance ID.
        :param source: The source account (defaults to transaction source).
        :return: A "revoke sponsorship" operation for a claimable balance.
        """
        return cls(
            revoke_sponsorship_type=RevokeSponsorshipType.CLAIMABLE_BALANCE,
            account_id=None,
            trustline=None,
            offer=None,
            data=None,
            claimable_balance_id=claimable_balance_id,
            signer=None,
            source=source,
        )

    @classmethod
    def revoke_signer_sponsorship(
        cls, account_id: str, signer_key: SignerKey, source: str = None
    ):
        """Create a "revoke sponsorship" operation for a signer.

        :param account_id: The account ID where the signer sponsorship is being removed from.
        :param signer_key: The signer whose sponsorship is being removed.
        :param source: The source account (defaults to transaction source).
        :return: A "revoke sponsorship" operation for a signer.
        """
        signer = Signer(account_id=account_id, signer_key=signer_key)
        return cls(
            revoke_sponsorship_type=RevokeSponsorshipType.SIGNER,
            account_id=None,
            trustline=None,
            offer=None,
            data=None,
            claimable_balance_id=None,
            signer=signer,
            source=source,
        )

    def _to_operation_body(self) -> Xdr.nullclass:
        body = Xdr.nullclass()
        body.type = Xdr.const.REVOKE_SPONSORSHIP
        if self.revoke_sponsorship_type == RevokeSponsorshipType.ACCOUNT:
            ledger_key = Xdr.nullclass()
            ledger_key.type = Xdr.const.ACCOUNT
            account = Xdr.nullclass()
            account.accountID = Keypair.from_public_key(
                self.account_id
            ).xdr_account_id()
            ledger_key.account = account
            revoke_sponsorship_op = Xdr.nullclass()
            revoke_sponsorship_op.type = Xdr.const.REVOKE_SPONSORSHIP_LEDGER_ENTRY
            revoke_sponsorship_op.ledgerKey = ledger_key
            body.revokeSponsorshipOp = revoke_sponsorship_op
            return body
        elif self.revoke_sponsorship_type == RevokeSponsorshipType.TRUSTLINE:
            ledger_key = Xdr.nullclass()
            ledger_key.type = Xdr.const.TRUSTLINE
            trust_line = Xdr.nullclass()
            trust_line.accountID = Keypair.from_public_key(
                self.trustline.account_id
            ).xdr_account_id()
            trust_line.asset = self.trustline.asset.to_xdr_object()
            ledger_key.trustLine = trust_line
            revoke_sponsorship_op = Xdr.nullclass()
            revoke_sponsorship_op.type = Xdr.const.REVOKE_SPONSORSHIP_LEDGER_ENTRY
            revoke_sponsorship_op.ledgerKey = ledger_key
            body.revokeSponsorshipOp = revoke_sponsorship_op
            return body
        elif self.revoke_sponsorship_type == RevokeSponsorshipType.OFFER:
            ledger_key = Xdr.nullclass()
            ledger_key.type = Xdr.const.OFFER
            offer = Xdr.nullclass()
            offer.sellerID = Keypair.from_public_key(
                self.offer.seller_id
            ).xdr_account_id()
            offer.offerID = self.offer.offer_id
            ledger_key.offer = offer
            revoke_sponsorship_op = Xdr.nullclass()
            revoke_sponsorship_op.type = Xdr.const.REVOKE_SPONSORSHIP_LEDGER_ENTRY
            revoke_sponsorship_op.ledgerKey = ledger_key
            body.revokeSponsorshipOp = revoke_sponsorship_op
            return body
        elif self.revoke_sponsorship_type == RevokeSponsorshipType.DATA:
            ledger_key = Xdr.nullclass()
            ledger_key.type = Xdr.const.DATA
            data = Xdr.nullclass()
            data.accountID = Keypair.from_public_key(
                self.data.account_id
            ).xdr_account_id()
            data.dataName = self.data.data_name.encode()
            ledger_key.data = data
            revoke_sponsorship_op = Xdr.nullclass()
            revoke_sponsorship_op.type = Xdr.const.REVOKE_SPONSORSHIP_LEDGER_ENTRY
            revoke_sponsorship_op.ledgerKey = ledger_key
            body.revokeSponsorshipOp = revoke_sponsorship_op
            return body
        elif self.revoke_sponsorship_type == RevokeSponsorshipType.CLAIMABLE_BALANCE:
            ledger_key = Xdr.nullclass()
            ledger_key.type = Xdr.const.CLAIMABLE_BALANCE
            claimable_balance_bytes = binascii.unhexlify(self.claimable_balance_id)
            claimable_balance = Xdr.nullclass()
            balance_id = ClaimableBalanceID.from_xdr(
                base64.b64encode(claimable_balance_bytes)
            )
            claimable_balance.balanceID = balance_id
            ledger_key.claimableBalance = claimable_balance
            revoke_sponsorship_op = Xdr.nullclass()
            revoke_sponsorship_op.type = Xdr.const.REVOKE_SPONSORSHIP_LEDGER_ENTRY
            revoke_sponsorship_op.ledgerKey = ledger_key
            body.revokeSponsorshipOp = revoke_sponsorship_op
            return body
        elif self.revoke_sponsorship_type == RevokeSponsorshipType.SIGNER:
            signer = Xdr.nullclass()
            signer.accountID = Keypair.from_public_key(
                self.signer.account_id
            ).xdr_account_id()
            signer.signerKey = self.signer.signer_key.to_xdr_object()
            revoke_sponsorship_op = Xdr.nullclass()
            revoke_sponsorship_op.type = Xdr.const.REVOKE_SPONSORSHIP_SIGNER
            revoke_sponsorship_op.signer = signer
            body.revokeSponsorshipOp = revoke_sponsorship_op
            return body
        else:
            raise ValueError(
                f"{self.revoke_sponsorship_type} is not a valid RevokeSponsorshipType."
            )

    @classmethod
    def from_xdr_object(
        cls, operation_xdr_object: Xdr.types.Operation
    ) -> "RevokeSponsorship":
        """Creates a :class:`RevokeSponsorship` object from an XDR Operation
        object.
        """
        source = Operation.get_source_from_xdr_obj(operation_xdr_object)
        op_type = operation_xdr_object.body.revokeSponsorshipOp.type
        if op_type == Xdr.const.REVOKE_SPONSORSHIP_LEDGER_ENTRY:
            ledger_key = operation_xdr_object.body.revokeSponsorshipOp.ledgerKey
            ledger_key_type = ledger_key.type
            if ledger_key_type == Xdr.const.ACCOUNT:
                account_id = StrKey.encode_ed25519_public_key(
                    ledger_key.account.accountID.ed25519
                )
                op = cls.revoke_account_sponsorship(account_id, source)
            elif ledger_key_type == Xdr.const.TRUSTLINE:
                account_id = StrKey.encode_ed25519_public_key(
                    ledger_key.trustLine.accountID.ed25519
                )
                asset = Asset.from_xdr_object(ledger_key.trustLine.asset)
                op = cls.revoke_trustline_sponsorship(account_id, asset, source)
            elif ledger_key_type == Xdr.const.OFFER:
                seller_id = StrKey.encode_ed25519_public_key(
                    ledger_key.offer.sellerID.ed25519
                )
                offer_id = ledger_key.offer.offerID
                op = cls.revoke_offer_sponsorship(seller_id, offer_id, source)
            elif ledger_key_type == Xdr.const.DATA:
                account_id = StrKey.encode_ed25519_public_key(
                    ledger_key.data.accountID.ed25519
                )
                data_name = ledger_key.data.dataName.decode()
                op = cls.revoke_data_sponsorship(account_id, data_name, source)
            elif ledger_key_type == Xdr.const.CLAIMABLE_BALANCE:
                balance_id = base64.b64decode(
                    ledger_key.claimableBalance.balanceID.to_xdr()
                )
                balance_id = binascii.hexlify(balance_id).decode()
                op = cls.revoke_claimable_balance_sponsorship(balance_id, source)
            else:
                raise ValueError(f"{ledger_key_type} is an unsupported LedgerKey type.")
        elif op_type == Xdr.const.REVOKE_SPONSORSHIP_SIGNER:
            account_id = StrKey.encode_ed25519_public_key(
                operation_xdr_object.body.revokeSponsorshipOp.signer.accountID.ed25519
            )
            signer_key = SignerKey.from_xdr_object(
                operation_xdr_object.body.revokeSponsorshipOp.signer.signerKey
            )
            op = cls.revoke_signer_sponsorship(account_id, signer_key, source)
        else:
            raise ValueError(f"{op_type} is an unsupported RevokeSponsorship type.")
        op._source_muxed = Operation.get_source_muxed_from_xdr_obj(operation_xdr_object)
        return op



import re
from typing import Optional, Dict

from .exceptions import AssetCodeInvalidError, AssetIssuerInvalidError, AttributeError
from .keypair import Keypair
from .xdr import Xdr
from .strkey import StrKey

__all__ = ["Asset"]


class Asset:
    """The :class:`Asset` object, which represents an asset and its
    corresponding issuer on the Stellar network.

    For more information about the formats used for asset codes and how issuers
    work on Stellar's network, see `Stellar's guide on assets`_.

    :param code: The asset code, in the formats specified in `Stellar's
        guide on assets`_.
    :param issuer: The account ID of the issuer. Note if the
        currency is the native currency (XLM (Lumens)), no issuer is necessary.
    :raises:
        | :exc:`AssetCodeInvalidError <stellar_sdk.exceptions.AssetCodeInvalidError>`: if ``code`` is invalid.
        | :exc:`AssetIssuerInvalidError <stellar_sdk.exceptions.AssetIssuerInvalidError>`: if ``issuer`` is not a valid ed25519 public key.

    .. _Stellar's guide on assets:
        https://www.stellar.org/developers/guides/concepts/assets.html
    """

    def __init__(self, code: str, issuer: Optional[str] = None) -> None:
        Asset.check_if_asset_code_is_valid(code)

        if code != "XLM" and issuer is None:
            raise AssetIssuerInvalidError(
                "The issuer cannot be `None` except for the native asset."
            )

        if issuer is not None and not StrKey.is_valid_ed25519_public_key(issuer):
            raise AssetIssuerInvalidError("The issuer should be a correct public key.")

        self.code: str = code
        self.issuer: Optional[str] = issuer
        self._type: str = self.guess_asset_type()

    @staticmethod
    def check_if_asset_code_is_valid(code: str) -> None:
        asset_code_re = re.compile(r"^[a-zA-Z0-9]{1,12}$")
        if not asset_code_re.match(code):
            raise AssetCodeInvalidError(
                "Asset code is invalid (maximum alphanumeric, 12 characters at max)."
            )

    @property
    def type(self) -> str:
        """Return the type of the asset, Can be one of following types: `native`, `credit_alphanum4` or `credit_alphanum12`

        :return: The type of the asset.
        """
        return self._type

    @type.setter
    def type(self, v) -> None:
        raise AttributeError("Asset type is immutable.")

    def guess_asset_type(self) -> str:
        """Return the type of the asset, Can be one of following types: `native`, `credit_alphanum4` or `credit_alphanum12`

        :return: The type of the asset.
        """
        if self.code == "XLM" and self.issuer is None:
            asset_type = "native"
        elif len(self.code) > 4:
            asset_type = "credit_alphanum12"
        else:
            asset_type = "credit_alphanum4"
        return asset_type

    def to_dict(self) -> dict:
        """Generate a dict for this object's attributes.

        :return: A dict representing an :class:`Asset`
        """
        rv: Dict[str, str] = {"type": self.type}
        if not self.is_native():
            rv["code"] = self.code
            rv["issuer"] = self.issuer
        return rv

    @staticmethod
    def native() -> "Asset":
        """Returns an asset object for the native asset.

        :return: An asset object for the native asset.
        """
        return Asset("XLM")

    def is_native(self) -> bool:
        """Return true if the :class:`Asset` is the native asset.

        :return: True if the Asset is native, False otherwise.
        """
        return self.issuer is None

    def to_xdr_object(self) -> Xdr.types.Asset:
        """Returns the xdr object for this asset.

        :return: XDR Asset object
        """
        if self.is_native():
            xdr_type = Xdr.const.ASSET_TYPE_NATIVE
            return Xdr.types.Asset(type=xdr_type)
        else:
            x = Xdr.nullclass()
            length = len(self.code)
            pad_length = 4 - length if length <= 4 else 12 - length
            x.assetCode = bytearray(self.code, "ascii") + b"\x00" * pad_length
            x.issuer = Keypair.from_public_key(self.issuer).xdr_account_id()
        if length <= 4:
            xdr_type = Xdr.const.ASSET_TYPE_CREDIT_ALPHANUM4
            return Xdr.types.Asset(type=xdr_type, alphaNum4=x)
        else:
            xdr_type = Xdr.const.ASSET_TYPE_CREDIT_ALPHANUM12
            return Xdr.types.Asset(type=xdr_type, alphaNum12=x)

    @classmethod
    def from_xdr_object(cls, asset_xdr_object: Xdr.types.Asset) -> "Asset":
        """Create a :class:`Asset` from an XDR Asset object.

        :param asset_xdr_object: The XDR Asset object.
        :return: A new :class:`Asset` object from the given XDR Asset object.
        """
        if asset_xdr_object.type == Xdr.const.ASSET_TYPE_NATIVE:
            return Asset.native()
        elif asset_xdr_object.type == Xdr.const.ASSET_TYPE_CREDIT_ALPHANUM4:
            issuer = StrKey.encode_ed25519_public_key(
                asset_xdr_object.alphaNum4.issuer.ed25519
            )
            code = asset_xdr_object.alphaNum4.assetCode.decode().rstrip("\x00")
        else:
            issuer = StrKey.encode_ed25519_public_key(
                asset_xdr_object.alphaNum12.issuer.ed25519
            )
            code = asset_xdr_object.alphaNum12.assetCode.decode().rstrip("\x00")
        return cls(code, issuer)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented  # pragma: no cover
        return self.code == other.code and self.issuer == other.issuer

    def __str__(self):
        return f"<Asset [code={self.code}, issuer={self.issuer}, type={self.type}]>"



from typing import Union

from ..call_builder.base_call_builder import BaseCallBuilder
from ..client.base_async_client import BaseAsyncClient
from ..client.base_sync_client import BaseSyncClient


class EffectsCallBuilder(BaseCallBuilder):
    """ Creates a new :class:`EffectsCallBuilder` pointed to server defined by horizon_url.
    Do not create this object directly, use :func:`stellar_sdk.server.Server.effects`.

    See `All Effects <https://www.stellar.org/developers/horizon/reference/endpoints/effects-all.html>`_

    :param horizon_url: Horizon server URL.
    :param client: The client instance used to send request.
    """

    def __init__(
        self, horizon_url: str, client: Union[BaseAsyncClient, BaseSyncClient]
    ) -> None:
        super().__init__(horizon_url, client)
        self.endpoint: str = "effects"

    def for_account(self, account_id: str) -> "EffectsCallBuilder":
        """This endpoint represents all effects that changed a given account. It will return relevant
        effects from the creation of the account to the current ledger.

        See `Effects for Account <https://www.stellar.org/developers/horizon/reference/endpoints/effects-all.html>`_

        :param account_id: account id, for example: `GDGQVOKHW4VEJRU2TETD6DBRKEO5ERCNF353LW5WBFW3JJWQ2BRQ6KDD`
        :return: this EffectCallBuilder instance
        """
        self.endpoint = f"accounts/{account_id}/effects"
        return self

    def for_ledger(self, sequence: Union[int, str]) -> "EffectsCallBuilder":
        """Effects are the specific ways that the ledger was changed by any operation.
        This endpoint represents all effects that occurred in the given ledger.

        See `Effects for Ledger <https://www.stellar.org/developers/horizon/reference/endpoints/effects-for-account.html>`_

        :param sequence: ledger sequence
        :return: this EffectCallBuilder instance
        """
        self.endpoint: str = f"ledgers/{sequence}/effects"
        return self

    def for_transaction(self, transaction_hash: str) -> "EffectsCallBuilder":
        """This endpoint represents all effects that occurred as a result of a given transaction.

        See `Effects for Transaction <https://www.stellar.org/developers/horizon/reference/endpoints/effects-for-transaction.html>`_

        :param transaction_hash: transaction hash
        :return: this EffectCallBuilder instance
        """
        self.endpoint: str = f"transactions/{transaction_hash}/effects"
        return self

    def for_operation(self, operation_id: Union[int, str]) -> "EffectsCallBuilder":
        """This endpoint represents all effects that occurred as a result of a given operation.

        See `Effects for Operation <https://www.stellar.org/developers/horizon/reference/endpoints/effects-for-operation.html>`_

        :param operation_id: operation ID
        :return: this EffectCallBuilder instance
        """
        self.endpoint: str = f"operations/{operation_id}/effects"
        return self



from .operation import Operation
from .utils import check_ed25519_public_key
from ..keypair import Keypair
from ..strkey import StrKey
from ..xdr import Xdr


class BeginSponsoringFutureReserves(Operation):
    """The :class:`BeginSponsoringFutureReserves` object, which represents a BeginSponsoringFutureReserves
    operation on Stellar's network.

    Establishes the is-sponsoring-future-reserves-for relationship between the source account and sponsoredID.
    See `Sponsored Reserves <https://developers.stellar.org/docs/glossary/sponsored-reserves/>_` for more information.

    See `Begin Sponsoring Future Reserves
    <https://developers.stellar.org/docs/start/list-of-operations/#begin-sponsoring-future-reserves>_`.

    Threshold: Medium

    :param sponsored_id: The sponsored account id.
    :param source: The source account (defaults to transaction source).
    """

    def __init__(self, sponsored_id: str, source: str = None) -> None:
        super().__init__(source)
        check_ed25519_public_key(sponsored_id)
        self.sponsored_id: str = sponsored_id

    @classmethod
    def type_code(cls) -> int:
        return Xdr.const.BEGIN_SPONSORING_FUTURE_RESERVES

    def _to_operation_body(self) -> Xdr.nullclass:
        sponsored_id = Keypair.from_public_key(self.sponsored_id).xdr_account_id()
        begin_sponsoring_future_reserves_op = Xdr.types.BeginSponsoringFutureReservesOp(
            sponsoredID=sponsored_id
        )
        body = Xdr.nullclass()
        body.type = Xdr.const.BEGIN_SPONSORING_FUTURE_RESERVES
        body.beginSponsoringFutureReservesOp = begin_sponsoring_future_reserves_op
        return body

    @classmethod
    def from_xdr_object(
        cls, operation_xdr_object: Xdr.types.Operation
    ) -> "BeginSponsoringFutureReserves":
        """Creates a :class:`BeginSponsoringFutureReserves` object from an XDR Operation
        object.
        """
        source = Operation.get_source_from_xdr_obj(operation_xdr_object)

        sponsored_id = StrKey.encode_ed25519_public_key(
            operation_xdr_object.body.beginSponsoringFutureReservesOp.sponsoredID.ed25519
        )
        op = cls(source=source, sponsored_id=sponsored_id)
        op._source_muxed = Operation.get_source_muxed_from_xdr_obj(operation_xdr_object)
        return op



import base64
import binascii

from .operation import Operation
from ..xdr import Xdr
from ..xdr.StellarXDR_type import ClaimableBalanceID


class ClaimClaimableBalance(Operation):
    """The :class:`ClaimClaimableBalance` object, which represents a ClaimClaimableBalance
    operation on Stellar's network.

    Claims a ClaimableBalanceEntry and adds the amount of asset on the entry to the source account.

    See `Claim Claimable Balance Documentation
    <https://developers.stellar.org/docs/start/list-of-operations/#claim-claimable-balance>_`.

    Threshold: Low

    :param balance_id: The claimable balance id to be claimed.
    :param source: The source account (defaults to transaction source).
    """

    def __init__(self, balance_id: str, source: str = None,) -> None:
        super().__init__(source)
        self.balance_id: str = balance_id

    @classmethod
    def type_code(cls) -> int:
        return Xdr.const.CLAIM_CLAIMABLE_BALANCE

    def _to_operation_body(self) -> Xdr.nullclass:
        body = Xdr.nullclass()
        body.type = Xdr.const.CLAIM_CLAIMABLE_BALANCE

        balance_id_bytes: bytes = binascii.unhexlify(self.balance_id)
        balance_id = ClaimableBalanceID.from_xdr(base64.b64encode(balance_id_bytes))
        claim_claimable_balance_op = Xdr.types.ClaimClaimableBalanceOp(
            balanceID=balance_id
        )

        body.claimClaimableBalanceOp = claim_claimable_balance_op
        return body

    @classmethod
    def from_xdr_object(
        cls, operation_xdr_object: Xdr.types.Operation
    ) -> "ClaimClaimableBalance":
        """Creates a :class:`ClaimClaimableBalance` object from an XDR Operation
        object.
        """
        source = Operation.get_source_from_xdr_obj(operation_xdr_object)
        balance_id = base64.b64decode(
            operation_xdr_object.body.claimClaimableBalanceOp.to_xdr()
        )
        balance_id = binascii.hexlify(balance_id).decode()
        op = cls(balance_id=balance_id, source=source)
        op._source_muxed = Operation.get_source_muxed_from_xdr_obj(operation_xdr_object)
        return op



from .operation import Operation
from ..xdr import Xdr


class Inflation(Operation):
    """The :class:`Inflation` object, which represents a
    Inflation operation on Stellar's network.

    This operation runs inflation.

    Threshold: Low

    :param str source: The source account (defaults to transaction source).

    """

    def __init__(self, source: str = None) -> None:
        super().__init__(source)

    @classmethod
    def type_code(cls) -> int:
        return Xdr.const.INFLATION

    def _to_operation_body(self) -> Xdr.nullclass:
        body = Xdr.nullclass()
        body.type = Xdr.const.INFLATION
        return body

    @classmethod
    def from_xdr_object(cls, operation_xdr_object: Xdr.types.Operation) -> "Inflation":
        """Creates a :class:`Inflation` object from an XDR Operation
        object.

        """
        source = Operation.get_source_from_xdr_obj(operation_xdr_object)
        op = cls(source)
        op._source_muxed = Operation.get_source_muxed_from_xdr_obj(operation_xdr_object)
        return op



from typing import Union

from ..call_builder.base_call_builder import BaseCallBuilder
from ..client.base_async_client import BaseAsyncClient
from ..client.base_sync_client import BaseSyncClient


class PaymentsCallBuilder(BaseCallBuilder):
    """ Creates a new :class:`PaymentsCallBuilder` pointed to server defined by horizon_url.
    Do not create this object directly, use :func:`stellar_sdk.server.Server.payments`.

    See `All Payments <https://www.stellar.org/developers/horizon/reference/endpoints/payments-all.html>`_

    :param horizon_url: Horizon server URL.
    :param client: The client instance used to send request.
    """

    def __init__(
        self, horizon_url: str, client: Union[BaseAsyncClient, BaseSyncClient]
    ) -> None:
        super().__init__(horizon_url, client)
        self.endpoint: str = "payments"

    def for_account(self, account_id: str) -> "PaymentsCallBuilder":
        """This endpoint responds with a collection of Payment operations where the given account
        was either the sender or receiver.

        See `Payments for Account <https://www.stellar.org/developers/horizon/reference/endpoints/payments-for-account.html>`_

        :param account_id: Account ID
        :return: current PaymentsCallBuilder instance
        """
        self.endpoint = f"accounts/{account_id}/payments"
        return self

    def for_ledger(self, sequence: Union[int, str]) -> "PaymentsCallBuilder":
        """This endpoint represents all payment operations that are part of a valid transactions in a given ledger.

        See `Payments for Ledger <https://www.stellar.org/developers/horizon/reference/endpoints/payments-for-ledger.html>`_

        :param sequence: Ledger sequence
        :return: current PaymentsCallBuilder instance
        """
        self.endpoint: str = f"ledgers/{sequence}/payments"
        return self

    def for_transaction(self, transaction_hash: str) -> "PaymentsCallBuilder":
        """This endpoint represents all payment operations that are part of a given transaction.

        See `Payments for Transaction <https://www.stellar.org/developers/horizon/reference/endpoints/payments-for-transaction.html>`_

        :param transaction_hash: Transaction hash
        :return: current PaymentsCallBuilder instance
        """
        self.endpoint: str = f"transactions/{transaction_hash}/payments"
        return self

    def include_failed(self, include_failed: bool) -> "PaymentsCallBuilder":
        """Adds a parameter defining whether to include failed transactions. By default only
        payments of successful transactions are returned.

        :param include_failed: Set to ``True`` to include payments of failed transactions.
        :return: current PaymentsCallBuilder instance
        """
        self._add_query_param("include_failed", include_failed)
        return self

    def join(self, join: str) -> "PaymentsCallBuilder":
        """join represents `join` param in queries, currently only supports `transactions`

        :param join: join represents `join` param in queries, currently only supports `transactions`
        :return: current OperationsCallBuilder instance
        """
        self._add_query_param("join", join)
        return self



from .xdr import Xdr
from .utils import best_rational_approximation

__all__ = ["Price"]


class Price:
    """Create a new price. Price in Stellar is represented as a fraction.

      :param n: numerator
      :param d: denominator
    """

    def __init__(self, n: int, d: int) -> None:
        self.n: int = n
        self.d: int = d

    @classmethod
    def from_raw_price(cls, price: str) -> "Price":
        """Create a :class:`Price` from the given str price.

        :param price: the str price. (ex. `'0.125'`)
        :return: A new :class:`Price` object from the given str price.
        :raises: :exc:`NoApproximationError <stellar_sdk.exceptions.NoApproximationError>`:
            if the approximation could not not be found.
        """
        best_r = best_rational_approximation(price)
        n = best_r["n"]
        d = best_r["d"]
        return cls(n, d)

    def to_xdr_object(self) -> Xdr.types.Price:
        """Returns the xdr object for this price object.

        :return: XDR Price object
        """
        return Xdr.types.Price(n=self.n, d=self.d)

    @classmethod
    def from_xdr_object(cls, price_xdr_object: Xdr.types.Price) -> "Price":
        """Create a :class:`Price` from an XDR Asset object.

        :param price_xdr_object: The XDR Price object.
        :return: A new :class:`Price` object from the given XDR Price object.
        """
        n = price_xdr_object.n
        d = price_xdr_object.d
        return cls(n, d)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented  # pragma: no cover
        return self.n == other.n and self.d == other.d

    def __str__(self):
        return f"<Price [n={self.n}, d={self.d}]>"



from .accounts_call_builder import AccountsCallBuilder
from .assets_call_builder import AssetsCallBuilder
from .base_call_builder import BaseCallBuilder
from .claimable_balances_call_builder import ClaimableBalancesCallBuilder
from .data_call_builder import DataCallBuilder
from .effects_call_builder import EffectsCallBuilder
from .fee_stats_call_builder import FeeStatsCallBuilder
from .ledgers_call_builder import LedgersCallBuilder
from .offers_call_builder import OffersCallBuilder
from .operations_call_builder import OperationsCallBuilder
from .orderbook_call_builder import OrderbookCallBuilder
from .paths_call_builder import PathsCallBuilder
from .payments_call_builder import PaymentsCallBuilder
from .root_call_builder import RootCallBuilder
from .trades_aggregation_call_builder import TradeAggregationsCallBuilder
from .trades_call_builder import TradesCallBuilder
from .transactions_call_builder import TransactionsCallBuilder
from .strict_receive_paths_call_builder import StrictReceivePathsCallBuilder
from .strict_send_paths_call_builder import StrictSendPathsCallBuilder

__all__ = [
    "AccountsCallBuilder",
    "AssetsCallBuilder",
    "BaseCallBuilder",
    "ClaimableBalancesCallBuilder",
    "DataCallBuilder",
    "EffectsCallBuilder",
    "FeeStatsCallBuilder",
    "LedgersCallBuilder",
    "OffersCallBuilder",
    "OperationsCallBuilder",
    "OrderbookCallBuilder",
    "PathsCallBuilder",
    "PaymentsCallBuilder",
    "RootCallBuilder",
    "StrictReceivePathsCallBuilder",
    "StrictSendPathsCallBuilder",
    "TradeAggregationsCallBuilder",
    "TradesCallBuilder",
    "TransactionsCallBuilder",
]



import warnings
from typing import Union

from ..asset import Asset
from ..call_builder import BaseCallBuilder
from ..client.base_async_client import BaseAsyncClient
from ..client.base_sync_client import BaseSyncClient


class OffersCallBuilder(BaseCallBuilder):
    """ Creates a new :class:`OffersCallBuilder` pointed to server defined by horizon_url.
    Do not create this object directly, use :func:`stellar_sdk.server.Server.offers`.

    See `Offer Details <https://www.stellar.org/developers/horizon/reference/endpoints/offer-details.html>`_
    See `Offers <https://www.stellar.org/developers/horizon/reference/endpoints/offers.html>`_

    :param horizon_url: Horizon server URL.
    :param client: The client instance used to send request.
    """

    def __init__(
        self, horizon_url: str, client: Union[BaseAsyncClient, BaseSyncClient]
    ) -> None:
        super().__init__(horizon_url, client)
        self.endpoint: str = "offers"

    def account(self, account_id):
        """Returns all offers where the given account is the seller.

        See `Offers for Account <https://www.stellar.org/developers/horizon/reference/endpoints/offers-for-account.html>`_

        :param account_id: Account ID
        :return: this OffersCallBuilder instance
        """
        warnings.warn(
            "Will be removed in future, use OffersCallBuilder.for_seller",
            DeprecationWarning,
        )
        self.endpoint = f"accounts/{account_id}/offers"
        return self

    def for_seller(self, seller: str):
        """Returns all offers where the given account is the seller.

        People on the Stellar network can make offers to buy or sell assets.
        This endpoint represents all the current offers, allowing
        filtering by `seller`, `selling_asset` or `buying_asset`.

        See `Offers <https://www.stellar.org/developers/horizon/reference/endpoints/offers.html>`_

        :param seller: Account ID of the offer creator
        :return: this OffersCallBuilder instance
        """
        self.endpoint: str = "offers"
        self._add_query_param("seller", seller)
        return self

    def for_buying(self, buying: Asset):
        """Returns all offers buying an asset.

        People on the Stellar network can make offers to buy or sell assets.
        This endpoint represents all the current offers, allowing
        filtering by `seller`, `selling_asset` or `buying_asset`.

        See `Offers <https://www.stellar.org/developers/horizon/reference/endpoints/offers.html>`_

        :param buying: The asset being bought.
        :return: this OffersCallBuilder instance
        """
        params = {
            "buying_asset_type": buying.type,
            "buying_asset_code": None if buying.is_native() else buying.code,
            "buying_asset_issuer": buying.issuer,
        }
        self._add_query_params(params)
        return self

    def for_selling(self, selling: Asset):
        """Returns all offers selling an asset.

        People on the Stellar network can make offers to buy or sell assets.
        This endpoint represents all the current offers, allowing
        filtering by `seller`, `selling_asset` or `buying_asset`.

        See `Offers <https://www.stellar.org/developers/horizon/reference/endpoints/offers.html>`_

        :param selling: The asset being sold.
        :return: this OffersCallBuilder instance
        """
        params = {
            "selling_asset_type": selling.type,
            "selling_asset_code": None if selling.is_native() else selling.code,
            "selling_asset_issuer": selling.issuer,
        }
        self._add_query_params(params)
        return self

    def offer(self, offer_id: Union[str, int]):
        """Returns information and links relating to a single offer.

        See `Offer Details <https://www.stellar.org/developers/horizon/reference/endpoints/offer-details.html>`_

        :param offer_id: Offer ID.
        :return: this OffersCallBuilder instance
        """
        self.endpoint = f"offers/{offer_id}"
        return self

    def for_sponsor(self, sponsor: str) -> "OffersCallBuilder":
        """Filtering offers where the given account is sponsoring the offer entry.

        See `Offer Details <https://developers.stellar.org/api/resources/offers/>`_

        :param sponsor: the sponsor id, for example: `GDGQVOKHW4VEJRU2TETD6DBRKEO5ERCNF353LW5WBFW3JJWQ2BRQ6KDD`
        :return: current OffersCallBuilder instance
        """
        self._add_query_param("sponsor", sponsor)
        return self



from .__version__ import __issues__
from .exceptions import ValueError
from .strkey import StrKey
from .xdr import Xdr

__all__ = ["SignerKey"]


class SignerKey:
    """The :class:`SignerKey` object, which represents an account signer key on Stellar's network.

    :param signer_key: The XDR signer object
    """

    def __init__(self, signer_key: Xdr.types.SignerKey) -> "None":
        self.signer_key: Xdr.types.SignerKey = signer_key

    @classmethod
    def ed25519_public_key(cls, account_id: str) -> "SignerKey":
        """Create ED25519 PUBLIC KEY Signer from account id.

        :param account_id: account id
        :return: ED25519 PUBLIC KEY Signer
        :raises:
            :exc:`Ed25519PublicKeyInvalidError <stellar_sdk.exceptions.Ed25519PublicKeyInvalidError>`: if ``account_id``
            is not a valid ed25519 public key.
        """
        signer_key = Xdr.types.SignerKey(
            Xdr.const.SIGNER_KEY_TYPE_ED25519,
            ed25519=StrKey.decode_ed25519_public_key(account_id),
        )

        return cls(signer_key)

    @classmethod
    def pre_auth_tx(cls, pre_auth_tx_hash: bytes) -> "SignerKey":
        """Create Pre AUTH TX Signer from the sha256 hash of a transaction,
        click `here <https://www.stellar.org/developers/guides/concepts/multi-sig.html#pre-authorized-transaction>`__ for more information.

        :param pre_auth_tx_hash: The sha256 hash of a transaction.
        :return: Pre AUTH TX Signer
        """
        signer_key = Xdr.types.SignerKey(
            Xdr.const.SIGNER_KEY_TYPE_PRE_AUTH_TX, preAuthTx=pre_auth_tx_hash
        )

        return cls(signer_key)

    @classmethod
    def sha256_hash(cls, sha256_hash: bytes) -> "SignerKey":
        """Create SHA256 HASH Signer from a sha256 hash of a preimage,
        click `here <https://www.stellar.org/developers/guides/concepts/multi-sig.html#hashx>`__ for more information.

        :param sha256_hash: a sha256 hash of a preimage
        :return: SHA256 HASH Signer
        """
        signer_key = Xdr.types.SignerKey(
            Xdr.const.SIGNER_KEY_TYPE_HASH_X, hashX=sha256_hash
        )
        return cls(signer_key)

    def to_xdr_object(self) -> Xdr.types.SignerKey:
        """Returns the xdr object for this SignerKey object.

        :return: XDR Signer object
        """
        return self.signer_key

    @classmethod
    def from_xdr_object(cls, signer_xdr_object: Xdr.types.Signer) -> "SignerKey":
        """Create a :class:`SignerKey` from an XDR SignerKey object.

        :param signer_xdr_object: The XDR SignerKey object.
        :return: A new :class:`SignerKey` object from the given XDR SignerKey object.
        """
        if signer_xdr_object.type == Xdr.const.SIGNER_KEY_TYPE_ED25519:
            account_id = StrKey.encode_ed25519_public_key(signer_xdr_object.ed25519)
            return cls.ed25519_public_key(account_id)
        elif signer_xdr_object.type == Xdr.const.SIGNER_KEY_TYPE_PRE_AUTH_TX:
            return cls.pre_auth_tx(signer_xdr_object.preAuthTx)
        elif signer_xdr_object.type == Xdr.const.SIGNER_KEY_TYPE_HASH_X:
            return cls.sha256_hash(signer_xdr_object.hashX)
        else:
            raise ValueError(
                f"This is an unknown signer type, please consider creating an issuer at {__issues__}."
            )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented  # pragma: no cover
        return self.to_xdr_object().to_xdr() == other.to_xdr_object().to_xdr()
